# SPDX-FileCopyrightText: 2009-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy
from bpy.types import Panel
from bpy.app.translations import contexts as i18n_contexts
from bl_ui.utils import PresetPanel
from bl_ui.properties_physics_common import (
    effector_weights_ui,
)


class FLUID_PT_presets(PresetPanel, Panel):
    bl_label = "Fluid Presets"
    preset_subdir = "fluid"
    preset_operator = "script.execute_preset"
    preset_add_operator = "fluid.preset_add"


class PhysicButtonsPanel:
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "physics"

    @staticmethod
    def check_domain_has_unbaked_guide(domain):
        return (
            domain.use_guide and not domain.has_cache_baked_guide and
            ((domain.guide_source == 'EFFECTOR') or
             (domain.guide_source == 'DOMAIN' and not domain.guide_parent))
        )

    @staticmethod
    def poll_fluid(context):
        ob = context.object
        if not ((ob and ob.type == 'MESH') and (context.fluid)):
            return False

        md = context.fluid
        return md and (context.fluid.fluid_type != 'NONE')

    @staticmethod
    def poll_fluid_domain(context):
        if not PhysicButtonsPanel.poll_fluid(context):
            return False

        md = context.fluid
        return md and (md.fluid_type == 'DOMAIN')

    @staticmethod
    def poll_gas_domain(context):
        if not PhysicButtonsPanel.poll_fluid(context):
            return False

        md = context.fluid
        if md and (md.fluid_type == 'DOMAIN'):
            domain = md.domain_settings
            return domain.domain_type == 'GAS'
        return False

    @staticmethod
    def poll_liquid_domain(context):
        if not PhysicButtonsPanel.poll_fluid(context):
            return False

        md = context.fluid
        if md and (md.fluid_type == 'DOMAIN'):
            domain = md.domain_settings
            return domain.domain_type == 'LIQUID'
        return False

    @staticmethod
    def poll_fluid_flow(context):
        if not PhysicButtonsPanel.poll_fluid(context):
            return False

        md = context.fluid
        return md and (md.fluid_type == 'FLOW')

    @staticmethod
    def poll_fluid_flow_outflow(context):
        if not PhysicButtonsPanel.poll_fluid_flow(context):
            return False

        md = context.fluid
        flow = md.flow_settings
        return (flow.flow_behavior == 'OUTFLOW')

    @staticmethod
    def poll_fluid_flow_liquid(context):
        if not PhysicButtonsPanel.poll_fluid_flow(context):
            return False

        md = context.fluid
        flow = md.flow_settings
        return (flow.flow_type == 'LIQUID')


class PHYSICS_PT_fluid(PhysicButtonsPanel, Panel):
    bl_label = "Fluid"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        ob = context.object
        return (ob and ob.type == 'MESH') and (context.engine in cls.COMPAT_ENGINES) and (context.fluid)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        if not bpy.app.build_options.fluid:
            col = layout.column(align=True)
            col.alignment = 'RIGHT'
            col.label(text="Built without Fluid modifier")
            return
        md = context.fluid

        layout.prop(md, "fluid_type")


class PHYSICS_PT_settings(PhysicButtonsPanel, Panel):
    bl_label = "Settings"
    bl_parent_id = "PHYSICS_PT_fluid"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        ob = context.object
        scene = context.scene

        if md.fluid_type == 'DOMAIN':
            domain = md.domain_settings

            is_baking_any = domain.is_cache_baking_any
            has_baked_data = domain.has_cache_baked_data

            row = layout.row()
            row.enabled = not is_baking_any and not has_baked_data
            row.prop(domain, "domain_type", expand=False)

            flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
            flow.enabled = not is_baking_any and not has_baked_data

            col = flow.column()
            col.enabled = not domain.has_cache_baked_guide
            col.prop(domain, "resolution_max", text="Resolution Divisions")
            col.prop(domain, "time_scale", text="Time Scale")

            col = flow.column()
            col.prop(domain, "use_adaptive_timesteps")
            sub = col.column(align=True)
            sub.active = domain.use_adaptive_timesteps
            sub.prop(domain, "cfl_condition", text="CFL Number")
            sub.prop(domain, "timesteps_max", text="Timesteps Maximum")
            sub.prop(domain, "timesteps_min", text="Minimum")

            col.separator()

            col = flow.column()
            if scene.use_gravity:
                sub = col.column()
                sub.enabled = False
                sub.prop(domain, "gravity", text="Scene Gravity", icon='SCENE_DATA')
            else:
                col.prop(domain, "gravity", text="Gravity")

            col = flow.column()
            if PhysicButtonsPanel.poll_gas_domain(context):
                col.prop(domain, "clipping", text="Empty Space")
            col.prop(domain, "delete_in_obstacle", text="Delete in Obstacle")

            if domain.cache_type == 'MODULAR':
                col.separator()
                label = ""

                # Deactivate bake operator if guides are enabled but not baked yet.
                note_flag = True
                if self.check_domain_has_unbaked_guide(domain):
                    note_flag = False
                    label = "Unbaked Guides: Bake Guides or disable them"
                elif not domain.cache_resumable and not label:
                    label = "Non Resumable Cache: Baking "
                    if PhysicButtonsPanel.poll_liquid_domain(context):
                        label += "mesh or particles will not be possible"
                    elif PhysicButtonsPanel.poll_gas_domain(context):
                        label += "noise will not be possible"
                    else:
                        label = ""

                if label:
                    info = layout.split()
                    note = info.row()
                    note.enabled = note_flag
                    note.alignment = 'RIGHT'
                    note.label(icon='INFO', text=label)

                split = layout.split()
                split.enabled = note_flag and ob.mode == 'OBJECT'

                bake_incomplete = (domain.cache_frame_pause_data < domain.cache_frame_end)
                if (
                        domain.cache_resumable and
                        domain.has_cache_baked_data and
                        not domain.is_cache_baking_data and
                        bake_incomplete
                ):
                    col = split.column()
                    col.operator("fluid.bake_data", text="Resume")
                    col = split.column()
                    col.operator("fluid.free_data", text="Free")
                elif domain.is_cache_baking_data and not domain.has_cache_baked_data:
                    split.enabled = False
                    split.operator("fluid.pause_bake", text="Baking Data - ESC to pause")
                elif not domain.has_cache_baked_data and not domain.is_cache_baking_data:
                    split.operator("fluid.bake_data", text="Bake Data")
                else:
                    split.operator("fluid.free_data", text="Free Data")

        elif md.fluid_type == 'FLOW':
            flow = md.flow_settings

            row = layout.row()
            row.prop(flow, "flow_type", expand=False)

            grid = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

            col = grid.column()
            col.prop(flow, "flow_behavior", expand=False)
            if flow.flow_behavior in {'INFLOW', 'OUTFLOW'}:
                col.prop(flow, "use_inflow")

            col.prop(flow, "subframes", text="Sampling Substeps")

            if not flow.flow_behavior == 'OUTFLOW' and flow.flow_type in {'SMOKE', 'BOTH', 'FIRE'}:

                if flow.flow_type in {'SMOKE', 'BOTH'}:
                    col.prop(flow, "smoke_color", text="Smoke Color")

                col = grid.column(align=True)
                col.prop(flow, "use_absolute", text="Absolute Density")

                if flow.flow_type in {'SMOKE', 'BOTH'}:
                    col.prop(flow, "temperature", text="Initial Temperature")
                    col.prop(flow, "density", text="Density")

                if flow.flow_type in {'FIRE', 'BOTH'}:
                    col.prop(flow, "fuel_amount", text="Fuel")

                col.separator()
                col.prop_search(flow, "density_vertex_group", ob, "vertex_groups", text="Vertex Group")

        elif md.fluid_type == 'EFFECTOR':
            effector_settings = md.effector_settings

            row = layout.row()
            row.prop(effector_settings, "effector_type")

            grid = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

            col = grid.column()
            col.prop(effector_settings, "subframes", text="Sampling Substeps")
            col.prop(effector_settings, "surface_distance", text="Surface Thickness")

            col = grid.column()

            col.prop(effector_settings, "use_effector", text="Use Effector")
            col.prop(effector_settings, "use_plane_init", text="Is Planar")

            if effector_settings.effector_type == 'GUIDE':
                col.prop(effector_settings, "velocity_factor", text="Velocity Factor")
                col.prop(effector_settings, "guide_mode", text="Guide Mode")


class PHYSICS_PT_borders(PhysicButtonsPanel, Panel):
    bl_label = "Border Collisions"
    bl_parent_id = "PHYSICS_PT_settings"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        domain = md.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        col = layout.column(align=True)
        col.enabled = not is_baking_any and not has_baked_data

        col.prop(domain, "use_collision_border_front")
        col.prop(domain, "use_collision_border_back")
        col.prop(domain, "use_collision_border_right")
        col.prop(domain, "use_collision_border_left")
        col.prop(domain, "use_collision_border_top")
        col.prop(domain, "use_collision_border_bottom")


class PHYSICS_PT_smoke(PhysicButtonsPanel, Panel):
    bl_label = "Gas"
    bl_parent_id = "PHYSICS_PT_fluid"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        domain = md.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_data

        col = flow.column(align=True)
        col.prop(domain, "alpha", text="Buoyancy Density")
        col.prop(domain, "beta", text="Heat")
        col = flow.column()
        col.prop(domain, "vorticity")


class PHYSICS_PT_smoke_dissolve(PhysicButtonsPanel, Panel):
    bl_label = "Dissolve"
    bl_translation_context = i18n_contexts.id_volume
    bl_parent_id = "PHYSICS_PT_smoke"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any

        self.layout.enabled = not is_baking_any
        self.layout.prop(md, "use_dissolve_smoke", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        domain = md.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_data

        layout.active = domain.use_dissolve_smoke

        col = flow.column()
        col.prop(domain, "dissolve_speed", text="Time")

        col = flow.column()
        col.prop(domain, "use_dissolve_smoke_log", text="Slow")


class PHYSICS_PT_fire(PhysicButtonsPanel, Panel):
    bl_label = "Fire"
    bl_parent_id = "PHYSICS_PT_smoke"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        domain = md.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_data

        col = flow.column()
        col.prop(domain, "burning_rate", text="Reaction Speed")
        row = col.row()
        sub = row.column(align=True)
        sub.prop(domain, "flame_smoke", text="Flame Smoke")
        sub.prop(domain, "flame_vorticity", text="Vorticity")

        col = flow.column(align=True)
        col.prop(domain, "flame_max_temp", text="Temperature Maximum")
        col.prop(domain, "flame_ignition", text="Minimum")
        row = col.row()
        row.prop(domain, "flame_smoke_color", text="Smoke Color")


class PHYSICS_PT_liquid(PhysicButtonsPanel, Panel):
    bl_label = "Liquid"
    bl_parent_id = "PHYSICS_PT_fluid"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_liquid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any

        self.layout.enabled = not is_baking_any
        self.layout.prop(md, "use_flip_particles", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        md = context.fluid
        domain = md.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        layout.enabled = not is_baking_any and not has_baked_data
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()
        col.prop(domain, "simulation_method", expand=False)
        if domain.simulation_method == 'FLIP':
            col.prop(domain, "flip_ratio", text="FLIP Ratio")
        col.prop(domain, "sys_particle_maximum", text="System Maximum")
        col = col.column(align=True)
        col.prop(domain, "particle_radius", text="Particle Radius")
        col.prop(domain, "particle_number", text="Sampling")
        col.prop(domain, "particle_randomness", text="Randomness")

        col = flow.column()
        col = col.column(align=True)
        col.prop(domain, "particle_max", text="Particles Maximum")
        col.prop(domain, "particle_min", text="Minimum")

        col.separator()

        col = col.column()
        col.prop(domain, "particle_band_width", text="Narrow Band Width")

        col = col.column()
        col.prop(domain, "use_fractions", text="Fractional Obstacles")
        sub = col.column()
        sub.active = domain.use_fractions
        sub.prop(domain, "fractions_distance", text="Obstacle Distance")
        sub.prop(domain, "fractions_threshold", text="Threshold")


class PHYSICS_PT_flow_source(PhysicButtonsPanel, Panel):
    bl_label = "Flow Source"
    bl_parent_id = "PHYSICS_PT_settings"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_flow(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        flow = context.fluid.flow_settings

        col = layout.column()
        col.prop(flow, "flow_source", expand=False, text="Flow Source")
        if flow.flow_source == 'PARTICLES':
            col.prop_search(flow, "particle_system", ob, "particle_systems", text="Particle System")

        grid = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = grid.column()
        if flow.flow_source == 'MESH':
            col.prop(flow, "use_plane_init", text="Is Planar")
            col.prop(flow, "surface_distance", text="Surface Emission")
            if flow.flow_type in {'SMOKE', 'BOTH', 'FIRE'}:
                col = grid.column()
                col.prop(flow, "volume_density", text="Volume Emission")

        if flow.flow_source == 'PARTICLES':
            col.prop(flow, "use_particle_size", text="Set Size")
            sub = col.column()
            sub.active = flow.use_particle_size
            sub.prop(flow, "particle_size")


class PHYSICS_PT_flow_initial_velocity(PhysicButtonsPanel, Panel):
    bl_label = "Initial Velocity"
    bl_parent_id = "PHYSICS_PT_settings"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_flow(context):
            return False

        if PhysicButtonsPanel.poll_fluid_flow_outflow(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid
        flow_smoke = md.flow_settings

        self.layout.prop(flow_smoke, "use_initial_velocity", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        flow = layout.grid_flow(row_major=False, columns=0, even_columns=True, even_rows=False, align=True)

        md = context.fluid
        flow_smoke = md.flow_settings

        flow.active = flow_smoke.use_initial_velocity

        col = flow.column()
        col.prop(flow_smoke, "velocity_factor")

        if flow_smoke.flow_source == 'MESH':
            col.prop(flow_smoke, "velocity_normal")
            # col.prop(flow_smoke, "velocity_random")
            col = flow.column()
            col.prop(flow_smoke, "velocity_coord")


class PHYSICS_PT_flow_texture(PhysicButtonsPanel, Panel):
    bl_label = "Texture"
    bl_parent_id = "PHYSICS_PT_settings"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_flow(context):
            return False

        if PhysicButtonsPanel.poll_fluid_flow_outflow(context):
            return False

        if PhysicButtonsPanel.poll_fluid_flow_liquid(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid
        flow_smoke = md.flow_settings

        self.layout.prop(flow_smoke, "use_texture", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        ob = context.object
        flow_smoke = context.fluid.flow_settings

        sub = flow.column()
        sub.active = flow_smoke.use_texture
        sub.prop(flow_smoke, "noise_texture")
        sub.prop(flow_smoke, "texture_map_type", text="Mapping")

        col = flow.column()
        sub = col.column()
        sub.active = flow_smoke.use_texture

        if flow_smoke.texture_map_type == 'UV':
            sub.prop_search(flow_smoke, "uv_layer", ob.data, "uv_layers")

        if flow_smoke.texture_map_type == 'AUTO':
            sub.prop(flow_smoke, "texture_size")

        sub.prop(flow_smoke, "texture_offset")


class PHYSICS_PT_adaptive_domain(PhysicButtonsPanel, Panel):
    bl_label = "Adaptive Domain"
    bl_parent_id = "PHYSICS_PT_settings"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        md = context.fluid
        domain = md.domain_settings
        # Effector guides require a fixed domain size
        if domain.use_guide and domain.guide_source == 'EFFECTOR':
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any

        self.layout.enabled = not is_baking_any and not has_baked_any
        self.layout.prop(md, "use_adaptive_domain", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        layout.active = domain.use_adaptive_domain

        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)
        flow.enabled = not is_baking_any and not has_baked_any

        col = flow.column()
        col.prop(domain, "additional_res", text="Add Resolution")
        col.prop(domain, "adapt_margin")

        col.separator()

        col = flow.column()
        col.prop(domain, "adapt_threshold", text="Threshold")


class PHYSICS_PT_noise(PhysicButtonsPanel, Panel):
    bl_label = "Noise"
    bl_parent_id = "PHYSICS_PT_smoke"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings
        is_baking_any = domain.is_cache_baking_any
        self.layout.enabled = not is_baking_any
        self.layout.prop(md, "use_noise", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        domain = context.fluid.domain_settings
        layout.active = domain.use_noise

        is_baking_any = domain.is_cache_baking_any
        has_baked_noise = domain.has_cache_baked_noise

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_noise

        col = flow.column()
        col.prop(domain, "noise_scale", text="Upres Factor")
        col.prop(domain, "noise_strength", text="Strength", text_ctxt=i18n_contexts.amount)

        col = flow.column()
        col.prop(domain, "noise_pos_scale", text="Scale")
        col.prop(domain, "noise_time_anim", text="Time")

        if domain.cache_type == 'MODULAR':
            col.separator()

            # Deactivate bake operator if data has not been baked yet.
            note_flag = True
            if domain.use_noise:
                label = ""
                if not domain.cache_resumable:
                    label = "Non Resumable Cache: Enable resumable option first"
                elif not domain.has_cache_baked_data:
                    label = "Unbaked Data: Bake Data first"

                if label:
                    info = layout.split()
                    note = info.row()
                    note_flag = False
                    note.enabled = note_flag
                    note.alignment = 'RIGHT'
                    note.label(icon='INFO', text=label)

            split = layout.split()
            split.enabled = domain.has_cache_baked_data and note_flag and ob.mode == 'OBJECT'

            bake_incomplete = (domain.cache_frame_pause_noise < domain.cache_frame_end)
            if domain.has_cache_baked_noise and not domain.is_cache_baking_noise and bake_incomplete:
                col = split.column()
                col.operator("fluid.bake_noise", text="Resume")
                col = split.column()
                col.operator("fluid.free_noise", text="Free")
            elif not domain.has_cache_baked_noise and domain.is_cache_baking_noise:
                split.enabled = False
                split.operator("fluid.pause_bake", text="Baking Noise - ESC to pause")
            elif not domain.has_cache_baked_noise and not domain.is_cache_baking_noise:
                split.operator("fluid.bake_noise", text="Bake Noise")
            else:
                split.operator("fluid.free_noise", text="Free Noise")


class PHYSICS_PT_mesh(PhysicButtonsPanel, Panel):
    bl_label = "Mesh"
    bl_parent_id = "PHYSICS_PT_liquid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_liquid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings
        is_baking_any = domain.is_cache_baking_any
        self.layout.enabled = not is_baking_any
        self.layout.prop(md, "use_mesh", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        domain = context.fluid.domain_settings
        layout.active = domain.use_mesh

        is_baking_any = domain.is_cache_baking_any
        has_baked_mesh = domain.has_cache_baked_mesh

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_mesh

        col = flow.column()

        col.prop(domain, "mesh_scale", text="Upres Factor")
        col.prop(domain, "mesh_particle_radius", text="Particle Radius")

        col = flow.column()
        col.prop(domain, "use_speed_vectors", text="Speed Vectors")

        col.separator()
        col.prop(domain, "mesh_generator", text="Mesh Generator")

        if domain.mesh_generator == 'IMPROVED':
            col = flow.column(align=True)
            col.prop(domain, "mesh_smoothen_pos", text="Smoothing Positive")
            col.prop(domain, "mesh_smoothen_neg", text="Negative")

            col = flow.column(align=True)
            col.prop(domain, "mesh_concave_upper", text="Concavity Upper")
            col.prop(domain, "mesh_concave_lower", text="Lower")

        # TODO(@sebbas): for now just interpolate any up-resolution grids, ie not sampling high-resolution grids
        # col.prop(domain, "highres_sampling", text="Flow Sampling:")

        if domain.cache_type == 'MODULAR':
            col.separator()

            # Deactivate bake operator if data has not been baked yet.
            note_flag = True
            if domain.use_mesh:
                label = ""
                if not domain.cache_resumable:
                    label = "Non Resumable Cache: Enable resumable option first"
                elif not domain.has_cache_baked_data:
                    label = "Unbaked Data: Bake Data first"

                if label:
                    info = layout.split()
                    note = info.row()
                    note_flag = False
                    note.enabled = note_flag
                    note.alignment = 'RIGHT'
                    note.label(icon='INFO', text=label)

            split = layout.split()
            split.enabled = domain.has_cache_baked_data and note_flag and ob.mode == 'OBJECT'

            bake_incomplete = (domain.cache_frame_pause_mesh < domain.cache_frame_end)
            if domain.has_cache_baked_mesh and not domain.is_cache_baking_mesh and bake_incomplete:
                col = split.column()
                col.operator("fluid.bake_mesh", text="Resume")
                col = split.column()
                col.operator("fluid.free_mesh", text="Free")
            elif not domain.has_cache_baked_mesh and domain.is_cache_baking_mesh:
                split.enabled = False
                split.operator("fluid.pause_bake", text="Baking Mesh - ESC to pause")
            elif not domain.has_cache_baked_mesh and not domain.is_cache_baking_mesh:
                split.operator("fluid.bake_mesh", text="Bake Mesh")
            else:
                split.operator("fluid.free_mesh", text="Free Mesh")


class PHYSICS_PT_particles(PhysicButtonsPanel, Panel):
    bl_label = "Particles"
    bl_parent_id = "PHYSICS_PT_liquid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_liquid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_particles = domain.has_cache_baked_particles
        using_particles = domain.use_spray_particles or domain.use_foam_particles or domain.use_bubble_particles

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any

        sndparticle_combined_export = domain.sndparticle_combined_export
        col = flow.column()
        row = col.row()
        row.enabled = sndparticle_combined_export in {'OFF', 'FOAM + BUBBLES'}
        row.prop(domain, "use_spray_particles", text="Spray")
        row.prop(domain, "use_foam_particles", text="Foam")
        row.prop(domain, "use_bubble_particles", text="Bubbles")

        col.separator()

        col.prop(domain, "sndparticle_combined_export")

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_particles
        flow.active = using_particles

        col = flow.column()
        col.prop(domain, "particle_scale", text="Upres Factor")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_potential_max_wavecrest", text="Wave Crest Potential Maximum")
        col.prop(domain, "sndparticle_potential_min_wavecrest", text="Minimum")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_potential_max_trappedair", text="Trapped Air Potential Maximum")
        col.prop(domain, "sndparticle_potential_min_trappedair", text="Minimum")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_potential_max_energy", text="Kinetic Energy Potential Maximum")
        col.prop(domain, "sndparticle_potential_min_energy", text="Minimum")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_potential_radius", text="Potential Radius")
        col.prop(domain, "sndparticle_update_radius", text="Particle Update Radius")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_sampling_wavecrest", text="Wave Crest Particle Sampling")
        col.prop(domain, "sndparticle_sampling_trappedair", text="Trapped Air Particle Sampling")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_life_max", text="Particle Life Maximum")
        col.prop(domain, "sndparticle_life_min", text="Minimum")
        col.separator()

        col = flow.column(align=True)
        col.prop(domain, "sndparticle_bubble_buoyancy", text="Bubble Buoyancy")
        col.prop(domain, "sndparticle_bubble_drag", text="Bubble Drag")
        col.separator()

        col = flow.column()
        col.prop(domain, "sndparticle_boundary", text="Particles in Boundary")

        if domain.cache_type == 'MODULAR':
            col.separator()

            # Deactivate bake operator if data has not been baked yet.
            note_flag = True
            if using_particles:
                label = ""
                if not domain.cache_resumable:
                    label = "Non Resumable Cache: Enable resumable option first"
                elif not domain.has_cache_baked_data:
                    label = "Unbaked Data: Bake Data first"

                if label:
                    info = layout.split()
                    note = info.row()
                    note_flag = False
                    note.enabled = note_flag
                    note.alignment = 'RIGHT'
                    note.label(icon='INFO', text=label)

            split = layout.split()
            split.enabled = (
                note_flag and
                ob.mode == 'OBJECT' and
                domain.has_cache_baked_data and (
                    domain.use_spray_particles or
                    domain.use_bubble_particles or
                    domain.use_foam_particles or
                    domain.use_tracer_particles
                )
            )

            bake_incomplete = (domain.cache_frame_pause_particles < domain.cache_frame_end)
            if domain.has_cache_baked_particles and not domain.is_cache_baking_particles and bake_incomplete:
                col = split.column()
                col.operator("fluid.bake_particles", text="Resume")
                col = split.column()
                col.operator("fluid.free_particles", text="Free")
            elif not domain.has_cache_baked_particles and domain.is_cache_baking_particles:
                split.enabled = False
                split.operator("fluid.pause_bake", text="Baking Particles - ESC to pause")
            elif not domain.has_cache_baked_particles and not domain.is_cache_baking_particles:
                split.operator("fluid.bake_particles", text="Bake Particles")
            else:
                split.operator("fluid.free_particles", text="Free Particles")


class PHYSICS_PT_diffusion(PhysicButtonsPanel, Panel):
    bl_label = "Diffusion"
    bl_parent_id = "PHYSICS_PT_liquid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        # Fluid diffusion only enabled for liquids (surface tension and viscosity not relevant for smoke)
        if not PhysicButtonsPanel.poll_liquid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings
        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any
        self.layout.enabled = not is_baking_any and not has_baked_any
        self.layout.prop(md, "use_diffusion", text="")

    def draw_header_preset(self, _context):
        FLUID_PT_presets.draw_panel_header(self.layout)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        layout.active = domain.use_diffusion

        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_any and not has_baked_data

        col = flow.column(align=True)
        col.prop(domain, "viscosity_base", text="Base")
        col.prop(domain, "viscosity_exponent", text="Exponent", slider=True)

        col = flow.column()
        col.prop(domain, "surface_tension", text="Surface Tension")


class PHYSICS_PT_viscosity(PhysicButtonsPanel, Panel):
    bl_label = "High Viscosity Solver"
    bl_parent_id = "PHYSICS_PT_diffusion"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        # Fluid viscosity only enabled for liquids
        if not PhysicButtonsPanel.poll_liquid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings
        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any
        self.layout.enabled = not is_baking_any and not has_baked_any
        self.layout.prop(md, "use_viscosity", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        layout.active = domain.use_viscosity

        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_any and not has_baked_data

        col = flow.column(align=True)
        col.prop(domain, "viscosity_value", text="Strength", text_ctxt=i18n_contexts.amount)


class PHYSICS_PT_guide(PhysicButtonsPanel, Panel):
    bl_label = "Guides"
    bl_parent_id = "PHYSICS_PT_fluid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw_header(self, context):
        md = context.fluid.domain_settings
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any

        self.layout.enabled = not is_baking_any
        self.layout.prop(md, "use_guide", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings

        layout.active = domain.use_guide

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_data

        col = flow.column()
        col.prop(domain, "guide_alpha", text="Weight")
        col.prop(domain, "guide_beta", text="Size")
        col.prop(domain, "guide_vel_factor", text="Velocity Factor")

        col = flow.column()
        col.prop(domain, "guide_source", text="Velocity Source")
        if domain.guide_source == 'DOMAIN':
            col.prop(domain, "guide_parent", text="Guide Parent")

        if domain.cache_type == 'MODULAR':
            col.separator()

            if domain.guide_source == 'EFFECTOR':
                split = layout.split()
                bake_incomplete = (domain.cache_frame_pause_guide < domain.cache_frame_end)
                if domain.has_cache_baked_guide and not domain.is_cache_baking_guide and bake_incomplete:
                    col = split.column()
                    col.operator("fluid.bake_guides", text="Resume")
                    col = split.column()
                    col.operator("fluid.free_guides", text="Free")
                elif not domain.has_cache_baked_guide and domain.is_cache_baking_guide:
                    split.enabled = False
                    split.operator("fluid.pause_bake", text="Baking Guides - ESC to pause")
                elif not domain.has_cache_baked_guide and not domain.is_cache_baking_guide:
                    split.operator("fluid.bake_guides", text="Bake Guides")
                else:
                    split.operator("fluid.free_guides", text="Free Guides")


class PHYSICS_PT_collections(PhysicButtonsPanel, Panel):
    bl_label = "Collections"
    bl_parent_id = "PHYSICS_PT_fluid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()
        col.prop(domain, "fluid_group", text="Flow")

        # col.prop(domain, "effector_group", text="Forces")
        col.prop(domain, "effector_group", text="Effector")


class PHYSICS_PT_cache(PhysicButtonsPanel, Panel):
    bl_label = "Cache"
    bl_parent_id = "PHYSICS_PT_fluid"
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout

        ob = context.object
        domain = context.fluid.domain_settings

        is_baking_any = domain.is_cache_baking_any
        has_baked_data = domain.has_cache_baked_data

        col = layout.column()
        col.prop(domain, "cache_directory", text="")
        col.enabled = not is_baking_any

        layout.use_property_split = True

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()
        row = col.row()
        row = row.column(align=True)
        row.prop(domain, "cache_frame_start", text="Frame Start")
        row.prop(domain, "cache_frame_end", text="End")
        row = col.row()
        row.enabled = domain.cache_type in {'MODULAR', 'ALL'}
        row.prop(domain, "cache_frame_offset", text="Offset")

        col.separator()

        col = flow.column()
        col.prop(domain, "cache_type", expand=False)

        row = col.row()
        row.enabled = not is_baking_any and not has_baked_data
        row.prop(domain, "cache_resumable", text="Resumable")

        if domain.cache_type == 'ALL':
            col.separator()
            split = layout.split()
            split.enabled = ob.mode == 'OBJECT'

            bake_incomplete = (domain.cache_frame_pause_data < domain.cache_frame_end)
            if (
                    domain.cache_resumable and
                    domain.has_cache_baked_data and
                    not domain.is_cache_baking_data and
                    bake_incomplete
            ):
                col = split.column()
                col.operator("fluid.bake_all", text="Resume")
                col = split.column()
                col.operator("fluid.free_all", text="Free")
            elif domain.is_cache_baking_data and not domain.has_cache_baked_data:
                split.enabled = False
                split.operator("fluid.pause_bake", text="Baking All - ESC to pause")
            elif not domain.has_cache_baked_data and not domain.is_cache_baking_data:
                split.operator("fluid.bake_all", text="Bake All")
            else:
                split.operator("fluid.free_all", text="Free All")


class PHYSICS_PT_export(PhysicButtonsPanel, Panel):
    bl_label = "Volumetric Data"
    bl_parent_id = "PHYSICS_PT_cache"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        md = context.fluid

        is_baking_any = domain.is_cache_baking_any
        has_baked_any = domain.has_cache_baked_any
        has_baked_data = domain.has_cache_baked_data
        has_baked_mesh = domain.has_cache_baked_mesh

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        flow.enabled = not is_baking_any and not has_baked_any

        col = flow.column()

        row = col.row()
        row.enabled = not is_baking_any and not has_baked_data
        row.prop(domain, "cache_data_format", text="Format")

        if domain.cache_data_format == 'OPENVDB':
            col.enabled = not is_baking_any and not has_baked_data
            col.prop(domain, "openvdb_cache_compress_type", text="Compression")

            col = flow.column()
            col.prop(domain, "openvdb_data_depth", text="Precision")

        if md.domain_settings.domain_type == 'LIQUID' and domain.use_mesh:
            row = col.row()
            row.enabled = not is_baking_any and not has_baked_mesh
            row.prop(domain, "cache_mesh_format", text="Meshes")

        # Only show the advanced panel to advanced users who know Mantaflow's birthday :)
        if bpy.app.debug_value == 3001:
            col = flow.column()
            col.prop(domain, "export_manta_script", text="Export Mantaflow Script")


class PHYSICS_PT_field_weights(PhysicButtonsPanel, Panel):
    bl_label = "Field Weights"
    bl_parent_id = "PHYSICS_PT_fluid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_fluid_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        domain = context.fluid.domain_settings
        effector_weights_ui(self, domain.effector_weights, 'SMOKE')


class PHYSICS_PT_viewport_display(PhysicButtonsPanel, Panel):
    bl_label = "Viewport Display"
    bl_parent_id = "PHYSICS_PT_fluid"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return (PhysicButtonsPanel.poll_fluid_domain(context))

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        domain = context.fluid.domain_settings

        col = flow.column(align=False)
        col.prop(domain, "display_thickness")

        sub = col.column()
        sub.prop(domain, "display_interpolation")

        if domain.use_color_ramp and domain.color_ramp_field == 'FLAGS':
            sub.enabled = False

        col = col.column()
        col.active = not domain.use_slice
        col.prop(domain, "slice_per_voxel")


class PHYSICS_PT_viewport_display_slicing(PhysicButtonsPanel, Panel):
    bl_label = "Slice"
    bl_parent_id = "PHYSICS_PT_viewport_display"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return (PhysicButtonsPanel.poll_fluid_domain(context))

    def draw_header(self, context):
        md = context.fluid.domain_settings

        self.layout.prop(md, "use_slice", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings

        layout.active = domain.use_slice

        col = layout.column()
        col.prop(domain, "slice_axis")
        col.prop(domain, "slice_depth")

        sub = col.column()
        sub.prop(domain, "show_gridlines")

        sub.active = domain.display_interpolation == 'CLOSEST' or domain.color_ramp_field == 'FLAGS'


class PHYSICS_PT_viewport_display_color(PhysicButtonsPanel, Panel):
    bl_label = "Grid Display"
    bl_parent_id = "PHYSICS_PT_viewport_display"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return (PhysicButtonsPanel.poll_fluid_domain(context))

    def draw_header(self, context):
        md = context.fluid.domain_settings

        self.layout.prop(md, "use_color_ramp", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        col = layout.column()
        col.active = domain.use_color_ramp
        col.prop(domain, "color_ramp_field")

        if not domain.color_ramp_field == 'FLAGS':
            col.prop(domain, "color_ramp_field_scale")

        col.use_property_split = False

        if domain.color_ramp_field[:3] != 'PHI' and domain.color_ramp_field not in {'FLAGS', 'PRESSURE'}:
            col = col.column()
            col.template_color_ramp(domain, "color_ramp", expand=True)


class PHYSICS_PT_viewport_display_debug(PhysicButtonsPanel, Panel):
    bl_label = "Vector Display"
    bl_parent_id = "PHYSICS_PT_viewport_display"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return (PhysicButtonsPanel.poll_fluid_domain(context))

    def draw_header(self, context):
        md = context.fluid.domain_settings

        self.layout.prop(md, "show_velocity", text="")

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=True)

        domain = context.fluid.domain_settings

        col = flow.column()
        col.active = domain.show_velocity
        col.prop(domain, "vector_display_type", text="Display As")

        if not domain.use_guide and domain.vector_field == 'GUIDE_VELOCITY':
            note = layout.split()
            note.label(icon='INFO', text="Enable Guides first! Defaulting to Fluid Velocity")

        if domain.vector_display_type == 'MAC':
            sub = col.column(heading="MAC Grid")
            sub.prop(domain, "vector_show_mac_x")
            sub.prop(domain, "vector_show_mac_y")
            sub.prop(domain, "vector_show_mac_z")
        else:
            col.prop(domain, "vector_scale_with_magnitude")

        col.prop(domain, "vector_field")
        col.prop(domain, "vector_scale")


class PHYSICS_PT_viewport_display_advanced(PhysicButtonsPanel, Panel):
    bl_label = "Advanced"
    bl_parent_id = "PHYSICS_PT_viewport_display"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        domain = context.fluid.domain_settings
        return PhysicButtonsPanel.poll_fluid_domain(context) and domain.use_slice and domain.show_gridlines

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings

        layout.active = domain.display_interpolation == 'CLOSEST' or domain.color_ramp_field == 'FLAGS'

        col = layout.column()
        col.prop(domain, "gridlines_color_field", text="Color Gridlines")

        if domain.gridlines_color_field == 'RANGE':
            if domain.use_color_ramp and domain.color_ramp_field != 'FLAGS':
                col.prop(domain, "gridlines_lower_bound")
                col.prop(domain, "gridlines_upper_bound")
                col.prop(domain, "gridlines_range_color")
                col.prop(domain, "gridlines_cell_filter")
            else:
                note = layout.split()
                if not domain.use_color_ramp:
                    note.label(icon='INFO', text="Enable Grid Display to use range highlighting!")
                else:
                    note.label(icon='INFO', text="Range highlighting for flags is not available!")


class PHYSICS_PT_fluid_domain_render(PhysicButtonsPanel, Panel):
    bl_label = "Render"
    bl_parent_id = "PHYSICS_PT_fluid"
    bl_options = {'DEFAULT_CLOSED'}
    COMPAT_ENGINES = {
        'BLENDER_RENDER',
        'BLENDER_EEVEE_NEXT',
        'BLENDER_WORKBENCH',
    }

    @classmethod
    def poll(cls, context):
        if not PhysicButtonsPanel.poll_gas_domain(context):
            return False

        return (context.engine in cls.COMPAT_ENGINES)

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        domain = context.fluid.domain_settings
        layout.prop(domain, "velocity_scale")


classes = (
    FLUID_PT_presets,
    PHYSICS_PT_fluid,
    PHYSICS_PT_settings,
    PHYSICS_PT_borders,
    PHYSICS_PT_adaptive_domain,
    PHYSICS_PT_smoke,
    PHYSICS_PT_smoke_dissolve,
    PHYSICS_PT_noise,
    PHYSICS_PT_fire,
    PHYSICS_PT_liquid,
    PHYSICS_PT_diffusion,
    PHYSICS_PT_viscosity,
    PHYSICS_PT_particles,
    PHYSICS_PT_mesh,
    PHYSICS_PT_guide,
    PHYSICS_PT_collections,
    PHYSICS_PT_cache,
    PHYSICS_PT_export,
    PHYSICS_PT_field_weights,
    PHYSICS_PT_flow_source,
    PHYSICS_PT_flow_initial_velocity,
    PHYSICS_PT_flow_texture,
    PHYSICS_PT_viewport_display,
    PHYSICS_PT_viewport_display_slicing,
    PHYSICS_PT_viewport_display_color,
    PHYSICS_PT_viewport_display_debug,
    PHYSICS_PT_viewport_display_advanced,
    PHYSICS_PT_fluid_domain_render,
)


if __name__ == "__main__":  # only for live edit.
    from bpy.utils import register_class
    for cls in classes:
        register_class(cls)



# SPDX-FileCopyrightText: 2006-2012 assimp team
# SPDX-FileCopyrightText: 2013 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

__all__ = (
    "parse",
    "data_types",
    "parse_version",
    "FBXElem",
)

from struct import unpack
import array
import zlib
from io import BytesIO

from . import data_types
from .fbx_utils_threading import MultiThreadedTaskConsumer

# at the end of each nested block, there is a NUL record to indicate
# that the sub-scope exists (i.e. to distinguish between P: and P : {})
_BLOCK_SENTINEL_LENGTH = ...
_BLOCK_SENTINEL_DATA = ...
read_fbx_elem_start = ...
_IS_BIG_ENDIAN = (__import__("sys").byteorder != 'little')
_HEAD_MAGIC = b'Kaydara FBX Binary\x20\x20\x00\x1a\x00'
from collections import namedtuple
FBXElem = namedtuple("FBXElem", ("id", "props", "props_type", "elems"))
del namedtuple


def read_uint(read):
    return unpack(b'<I', read(4))[0]


def read_ubyte(read):
    return unpack(b'B', read(1))[0]


def read_string_ubyte(read):
    size = read_ubyte(read)
    data = read(size)
    return data


def read_array_params(read):
    return unpack(b'<III', read(12))


def read_elem_start32(read):
    end_offset, prop_count, _prop_length, elem_id_size = unpack(b'<IIIB', read(13))
    elem_id = read(elem_id_size) if elem_id_size else b""
    return end_offset, prop_count, elem_id


def read_elem_start64(read):
    end_offset, prop_count, _prop_length, elem_id_size = unpack(b'<QQQB', read(25))
    elem_id = read(elem_id_size) if elem_id_size else b""
    return end_offset, prop_count, elem_id


def _create_array(data, length, array_type, array_stride, array_byteswap):
    """Create an array from FBX data."""
    # If size of the data does not match the expected size of the array, then something is wrong with the code or the
    # FBX file.
    assert(length * array_stride == len(data))

    data_array = array.array(array_type, data)
    if array_byteswap and _IS_BIG_ENDIAN:
        data_array.byteswap()
    return data_array


def _decompress_and_insert_array(elem_props_data, index_to_set, compressed_array_args):
    """Decompress array data and insert the created array into the FBX tree being parsed.

    This is usually called from a separate thread to the main thread."""
    compressed_data, length, array_type, array_stride, array_byteswap = compressed_array_args

    # zlib.decompress releases the Global Interpreter Lock, so another thread can run code while waiting for the
    # decompression to complete.
    data = zlib.decompress(compressed_data, bufsize=length * array_stride)

    # Create and insert the array into the parsed FBX hierarchy.
    elem_props_data[index_to_set] = _create_array(data, length, array_type, array_stride, array_byteswap)


def unpack_array(read, array_type, array_stride, array_byteswap):
    """Unpack an array from an FBX file being parsed.

    If the array data is compressed, the compressed data is combined with the other arguments into a tuple to prepare
    for decompressing on a separate thread if possible.

    If the array data is not compressed, the array is created.

    Returns (tuple, True) or (array, False)."""
    length, encoding, comp_len = read_array_params(read)

    data = read(comp_len)

    if encoding == 1:
        # Array data requires decompression, which is done in a separate thread if possible.
        return (data, length, array_type, array_stride, array_byteswap), True
    else:
        return _create_array(data, length, array_type, array_stride, array_byteswap), False


read_array_dict = {
    b'b'[0]: lambda read: unpack_array(read, data_types.ARRAY_BOOL, 1, False),     # bool
    b'c'[0]: lambda read: unpack_array(read, data_types.ARRAY_BYTE, 1, False),     # ubyte
    b'i'[0]: lambda read: unpack_array(read, data_types.ARRAY_INT32, 4, True),     # int
    b'l'[0]: lambda read: unpack_array(read, data_types.ARRAY_INT64, 8, True),     # long
    b'f'[0]: lambda read: unpack_array(read, data_types.ARRAY_FLOAT32, 4, False),  # float
    b'd'[0]: lambda read: unpack_array(read, data_types.ARRAY_FLOAT64, 8, False),  # double
}

read_data_dict = {
    b'Z'[0]: lambda read: unpack(b'<b', read(1))[0],  # byte
    b'Y'[0]: lambda read: unpack(b'<h', read(2))[0],  # 16 bit int
    b'B'[0]: lambda read: unpack(b'?', read(1))[0],   # 1 bit bool (yes/no)
    b'C'[0]: lambda read: unpack(b'<c', read(1))[0],  # char
    b'I'[0]: lambda read: unpack(b'<i', read(4))[0],  # 32 bit int
    b'F'[0]: lambda read: unpack(b'<f', read(4))[0],  # 32 bit float
    b'D'[0]: lambda read: unpack(b'<d', read(8))[0],  # 64 bit float
    b'L'[0]: lambda read: unpack(b'<q', read(8))[0],  # 64 bit int
    b'R'[0]: lambda read: read(read_uint(read)),      # binary data
    b'S'[0]: lambda read: read(read_uint(read)),      # string data
}


# FBX 7500 (aka FBX2016) introduces incompatible changes at binary level:
#   * The NULL block marking end of nested stuff switches from 13 bytes long to 25 bytes long.
#   * The FBX element metadata (end_offset, prop_count and prop_length) switch from uint32 to uint64.
def init_version(fbx_version):
    global _BLOCK_SENTINEL_LENGTH, _BLOCK_SENTINEL_DATA, read_fbx_elem_start

    _BLOCK_SENTINEL_LENGTH = ...
    _BLOCK_SENTINEL_DATA = ...

    if fbx_version < 7500:
        _BLOCK_SENTINEL_LENGTH = 13
        read_fbx_elem_start = read_elem_start32
    else:
        _BLOCK_SENTINEL_LENGTH = 25
        read_fbx_elem_start = read_elem_start64
    _BLOCK_SENTINEL_DATA = (b'\0' * _BLOCK_SENTINEL_LENGTH)


def read_elem(read, tell, use_namedtuple, decompress_array_func, tell_file_offset=0):
    # [0] the offset at which this block ends
    # [1] the number of properties in the scope
    # [2] the length of the property list
    # [3] elem name length
    # [4] elem name of the scope/key
    # read_fbx_elem_start does not return [2] because we don't use it and does not return [3] because it is only used to
    # get [4].
    end_offset, prop_count, elem_id = read_fbx_elem_start(read)
    if end_offset == 0:
        return None

    elem_props_type = bytearray(prop_count)  # elem property types
    elem_props_data = [None] * prop_count    # elem properties (if any)
    elem_subtree = []                        # elem children (if any)

    for i in range(prop_count):
        data_type = read(1)[0]
        if data_type in read_array_dict:
            val, needs_decompression = read_array_dict[data_type](read)
            if needs_decompression:
                # Array decompression releases the GIL, so can be multithreaded (if possible on the current system) for
                # performance.
                # After decompressing, the array is inserted into elem_props_data[i].
                decompress_array_func(elem_props_data, i, val)
            else:
                elem_props_data[i] = val
        else:
            elem_props_data[i] = read_data_dict[data_type](read)
        elem_props_type[i] = data_type

    pos = tell()
    local_end_offset = end_offset - tell_file_offset

    if pos < local_end_offset:
        # The default BufferedReader used when `open()`-ing files in 'rb' mode has to get the raw stream position from
        # the OS every time its tell() function is called. This is about 10 times slower than the tell() function of
        # BytesIO objects, so reading chunks of bytes from the file into memory at once and exposing them through
        # BytesIO can give better performance. We know the total size of each element's subtree so can read entire
        # subtrees into memory at a time.
        # The "Objects" element's subtree, however, usually makes up most of the file, so we specifically avoid reading
        # all its sub-elements into memory at once to reduce memory requirements at the cost of slightly worse
        # performance when memory is not a concern.
        # If we're currently reading directly from the opened file, then tell_file_offset will be zero.
        if tell_file_offset == 0 and elem_id != b"Objects":
            block_bytes_remaining = local_end_offset - pos

            # Read the entire subtree
            sub_elem_bytes = read(block_bytes_remaining)
            num_bytes_read = len(sub_elem_bytes)
            if num_bytes_read != block_bytes_remaining:
                raise IOError("failed to read complete nested block, expected %i bytes, but only got %i"
                              % (block_bytes_remaining, num_bytes_read))

            # BytesIO provides IO API for reading bytes in memory, so we can use the same code as reading bytes directly
            # from a file.
            f = BytesIO(sub_elem_bytes)
            tell = f.tell
            read = f.read
            # The new `tell` function starts at zero and is offset by `pos` bytes from the start of the file.
            start_sub_pos = 0
            tell_file_offset = pos
            sub_tree_end = block_bytes_remaining - _BLOCK_SENTINEL_LENGTH
        else:
            # The `tell` function is unchanged, so starts at the value returned by `tell()`, which is still `pos`
            # because no reads have been made since then.
            start_sub_pos = pos
            sub_tree_end = local_end_offset - _BLOCK_SENTINEL_LENGTH

        sub_pos = start_sub_pos
        while sub_pos < sub_tree_end:
            elem_subtree.append(read_elem(read, tell, use_namedtuple, decompress_array_func, tell_file_offset))
            sub_pos = tell()

        # At the end of each subtree there should be a sentinel (an empty element with all bytes set to zero).
        if read(_BLOCK_SENTINEL_LENGTH) != _BLOCK_SENTINEL_DATA:
            raise IOError("failed to read nested block sentinel, "
                          "expected all bytes to be 0")

        # Update `pos` for the number of bytes that have been read.
        pos += (sub_pos - start_sub_pos) + _BLOCK_SENTINEL_LENGTH

    if pos != local_end_offset:
        raise IOError("scope length not reached, something is wrong")

    args = (elem_id, elem_props_data, elem_props_type, elem_subtree)
    return FBXElem(*args) if use_namedtuple else args


def parse_version(fn):
    """
    Return the FBX version,
    if the file isn't a binary FBX return zero.
    """
    with open(fn, 'rb') as f:
        read = f.read

        if read(len(_HEAD_MAGIC)) != _HEAD_MAGIC:
            return 0

        return read_uint(read)


def parse(fn, use_namedtuple=True):
    root_elems = []

    multithread_decompress_array_cm = MultiThreadedTaskConsumer.new_cpu_bound_cm(_decompress_and_insert_array)
    with open(fn, 'rb') as f, multithread_decompress_array_cm as decompress_array_func:
        read = f.read
        tell = f.tell

        if read(len(_HEAD_MAGIC)) != _HEAD_MAGIC:
            raise IOError("Invalid header")

        fbx_version = read_uint(read)
        init_version(fbx_version)

        while True:
            elem = read_elem(read, tell, use_namedtuple, decompress_array_func)
            if elem is None:
                break
            root_elems.append(elem)

    args = (b'', [], bytearray(0), root_elems)
    return FBXElem(*args) if use_namedtuple else args, fbx_version



# SPDX-FileCopyrightText: 2010-2022 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

import bpy

from ...base_rig import BaseRig

from ...utils.naming import make_derived_name
from ...utils.bones import set_bone_widget_transform
from ...utils.widgets import layout_widget_dropdown, create_registered_widget
from ...utils.widgets_basic import create_pivot_widget
from ...utils.switch_parent import SwitchParentBuilder


class Rig(BaseRig):
    """ A rig providing a rotation pivot control that can be moved. """

    class CtrlBones(BaseRig.CtrlBones):
        master: str
        pivot: str

    class MchBones(BaseRig.MchBones):
        pass

    bones: BaseRig.ToplevelBones[
        str,
        'Rig.CtrlBones',
        'Rig.MchBones',
        str
    ]

    make_control: bool
    make_pivot: bool
    make_deform: bool

    def find_org_bones(self, pose_bone) -> str:
        return pose_bone.name

    def initialize(self):
        self.make_control = self.params.make_extra_control
        self.make_pivot = self.params.make_control or not self.make_control
        self.make_deform = self.params.make_extra_deform

    def generate_bones(self):
        org = self.bones.org

        if self.make_control:
            self.bones.ctrl.master = name = self.copy_bone(org, make_derived_name(org, 'ctrl'), parent=True)

            if self.make_pivot:
                self.bones.ctrl.pivot = self.copy_bone(org, make_derived_name(org, 'ctrl', '_pivot'))

            if self.params.make_parent_switch:
                self.build_parent_switch(name)

            if self.params.register_parent:
                self.register_parent(name, self.get_parent_tags())

        else:
            self.bones.ctrl.pivot = self.copy_bone(org, make_derived_name(org, 'ctrl'), parent=True)

        if self.make_deform:
            self.bones.deform = self.copy_bone(org, make_derived_name(org, 'def'), bbone=True)

    def build_parent_switch(self, master_name: str):
        pbuilder = SwitchParentBuilder(self.generator)

        org_parent = self.get_bone_parent(self.bones.org)
        parents = [org_parent] if org_parent else []

        pbuilder.build_child(
            self, master_name,
            context_rig=self.rigify_parent, allow_self=True,
            prop_name="Parent ({})".format(master_name),
            extra_parents=parents, select_parent=org_parent,
            controls=lambda: self.bones.ctrl.flatten()
        )

    def get_parent_tags(self):
        tags = {t.strip() for t in self.params.register_parent_tags.split(',')}

        if self.params.make_parent_switch:
            tags.add('child')

        tags.discard('')
        return tags

    def register_parent(self, master_name: str, tags: set[str]):
        pbuilder = SwitchParentBuilder(self.generator)

        inject = self.rigify_parent if 'injected' in tags else None

        pbuilder.register_parent(
            self, self.bones.org, name=master_name,
            inject_into=inject, tags=tags
        )

    def parent_bones(self):
        ctrl = self.bones.ctrl

        if self.make_pivot:
            if self.make_control:
                self.set_bone_parent(ctrl.pivot, ctrl.master, use_connect=False)

            self.set_bone_parent(self.bones.org, ctrl.pivot, use_connect=False)

        else:
            self.set_bone_parent(self.bones.org, ctrl.master, use_connect=False)

        if self.make_deform:
            self.set_bone_parent(self.bones.deform, self.bones.org, use_connect=False)

    def configure_bones(self):
        org = self.bones.org
        ctrl = self.bones.ctrl
        main_ctl = ctrl.master if self.make_control else ctrl.pivot

        self.copy_bone_properties(org, main_ctl, ui_controls=True)

    def rig_bones(self):
        if self.make_pivot:
            self.make_constraint(
                self.bones.org, 'COPY_LOCATION', self.bones.ctrl.pivot,
                space='LOCAL', invert_xyz=(True,) * 3
            )

    def generate_widgets(self):
        if self.make_pivot:
            create_pivot_widget(self.obj, self.bones.ctrl.pivot, square=True, axis_size=2.0)

        if self.make_control:
            set_bone_widget_transform(self.obj, self.bones.ctrl.master, self.bones.org)

            create_registered_widget(self.obj, self.bones.ctrl.master,
                                     self.params.pivot_master_widget_type or 'cube')

    @classmethod
    def add_parameters(cls, params):
        params.make_control = bpy.props.BoolProperty(
            name="Control",
            default=True,
            description="Create a control bone for the copy"
        )

        params.pivot_master_widget_type = bpy.props.StringProperty(
            name="Widget Type",
            default='cube',
            description="Choose the type of the widget to create"
        )

        params.make_parent_switch = bpy.props.BoolProperty(
            name="Switchable Parent",
            default=False,
            description="Allow switching the parent of the master control"
        )

        params.register_parent = bpy.props.BoolProperty(
            name="Register Parent",
            default=False,
            description="Register the control as a switchable parent candidate"
        )

        params.register_parent_tags = bpy.props.StringProperty(
            name="Parent Tags",
            default="",
            description="Comma-separated tags to use for the registered parent"
        )

        params.make_extra_control = bpy.props.BoolProperty(
            name="Extra Control",
            default=False,
            description="Create an optional control"
        )

        params.make_extra_deform = bpy.props.BoolProperty(
            name="Extra Deform",
            default=False,
            description="Create an optional deform bone"
        )

    @classmethod
    def parameters_ui(cls, layout, params):
        r = layout.row()
        r.prop(params, "make_extra_control", text="Master Control")

        if params.make_extra_control:
            layout_widget_dropdown(layout, params, "pivot_master_widget_type")

            layout.prop(params, "make_parent_switch")
            layout.prop(params, "register_parent")

            r = layout.row()
            r.active = params.register_parent
            r.prop(params, "register_parent_tags", text="Tags")

            layout.prop(params, "make_control", text="Pivot Control")

        r = layout.row()
        r.prop(params, "make_extra_deform", text="Deform Bone")


def create_sample(obj):
    """ Create a sample metarig for this rig type.
    """
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('pivot')
    bone.head[:] = 0.0000, 0.0000, 0.0000
    bone.tail[:] = 0.0000, 0.5000, 0.0000
    bone.roll = 0.0000
    bone.use_connect = False
    bones['pivot'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['pivot']]
    pbone.rigify_type = 'basic.pivot'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone
        if bcoll := arm.collections.active:
            bcoll.assign(bone)

    return bones



# SPDX-FileCopyrightText: 2012-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Some misc utilities...

__all__ = (
    "I18n",
    "I18nMessage",
    "I18nMessages",
    "enable_addons",
    "find_best_isocode_matches",
    "get_po_files_from_dir",
    "list_po_dir",
)

import collections
import os
import platform
import re
import struct
import tempfile
import time

from bl_i18n_utils import (
    settings,
    utils_rtl,
)


##### Misc Utils #####
_valid_po_path_re = re.compile(r"^\S+:[0-9]+$")


def is_valid_po_path(path):
    return bool(_valid_po_path_re.match(path))


def get_best_similar(data):
    import difflib
    key, use_similar, similar_pool = data

    # try to find some close key in existing messages...
    # Optimized code inspired by difflib.get_close_matches (as we only need the best match).
    # We also consider to never make a match when len differs more than -len_key / 2, +len_key * 2 (which is valid
    # as long as use_similar is not below ~0.7).
    # Gives an overall ~20% of improvement!

    # tmp = difflib.get_close_matches(key[1], similar_pool, n=1, cutoff=use_similar)
    # if tmp:
    #     tmp = tmp[0]
    tmp = None
    s = difflib.SequenceMatcher()
    s.set_seq2(key[1])
    len_key = len(key[1])
    min_len = len_key // 2
    max_len = len_key * 2
    for x in similar_pool:
        if min_len < len(x) < max_len:
            s.set_seq1(x)
            if s.real_quick_ratio() >= use_similar and s.quick_ratio() >= use_similar:
                sratio = s.ratio()
                if sratio >= use_similar:
                    tmp = x
                    use_similar = sratio
    return key, tmp


_locale_explode_re = re.compile(r"^([a-z]{2,})(?:_([A-Za-z]{2,}))?(?:@([a-z]{2,}))?$")


def locale_explode(locale):
    """Copies behavior of `BLT_lang_locale_explode`, keep them in sync."""
    ret = (None, None, None, None, None)
    m = _locale_explode_re.match(locale)
    if m:
        lang, country, variant = m.groups()
        return (
            lang, country, variant,
            "{:s}_{:s}".format(lang, country) if country else None,
            "{:s}@{:s}".format(lang, variant) if variant else None
        )
    try:
        import bpy.app.translations as bpy_translations
        assert ret == bpy_translations.locale_explode(locale)
    except ModuleNotFoundError:
        pass

    return ret


def locale_match(loc1, loc2):
    """
    Return:
        -n if loc1 is a subtype of loc2 (e.g. ``fr_FR`` is a subtype of ``fr``).
        +n if loc2 is a subtype of loc1.
        n becomes smaller when both locales are more similar
        ... (e.g. (``sr, sr_SR``) are more similar than (``sr, sr_SR@latin``)).
        0 if they are exactly the same.
        ... (Ellipsis) if they cannot match!
    Note: We consider that ``sr_SR@latin`` is a subtype of ``sr@latin``, ``sr_SR`` and ``sr``,
          but ``sr_SR`` and ``sr@latin`` won't match (will return ...)!
    Note: About similarity, diff in variants are more important than diff in countries, currently here are the cases:
            (``sr, sr_SR``)             -> 1
            (``sr@latin, sr_SR@latin``) -> 1
            (``sr, sr@latin``)          -> 2
            (``sr_SR, sr_SR@latin``)    -> 2
            (``sr, sr_SR@latin``)       -> 3
    """
    if loc1 == loc2:
        return 0
    l1, c1, v1, *_1 = locale_explode(loc1)
    l2, c2, v2, *_2 = locale_explode(loc2)

    if l1 == l2:
        if c1 == c2:
            if v1 == v2:
                return 0
            elif v2 is None:
                return -2
            elif v1 is None:
                return 2
            return ...
        elif c2 is None:
            if v1 == v2:
                return -1
            elif v2 is None:
                return -3
            return ...
        elif c1 is None:
            if v1 == v2:
                return 1
            elif v1 is None:
                return 3
            return ...
    return ...


def find_best_isocode_matches(uid, iso_codes):
    """
    Return an ordered tuple of elements in iso_codes that can match the given uid, from most similar to lesser ones.
    """
    tmp = ((e, locale_match(e, uid)) for e in iso_codes)
    return tuple(e[0] for e in sorted((e for e in tmp if e[1] is not ... and e[1] >= 0), key=lambda e: e[1]))


def get_po_files_from_dir(root_dir, langs=set()):
    """
    Yield tuples (uid, po_path) of translations for each po file found in the given directory, which should be either
    a directory containing po files using language uid's as names (e.g. fr.po, es_ES.po, etc.), or
    a directory containing directories which names are language uids, and containing po files of the same names.
    """
    found_uids = set()
    for p in os.listdir(root_dir):
        uid = None
        po_file = os.path.join(root_dir, p)
        print(p)
        if p.endswith(".po") and os.path.isfile(po_file):
            uid = p[:-3]
            if langs and uid not in langs:
                continue
        elif os.path.isdir(p):
            uid = p
            if langs and uid not in langs:
                continue
            po_file = os.path.join(root_dir, p, p + ".po")
            if not os.path.isfile(po_file):
                continue
        else:
            continue
        if uid in found_uids:
            print("WARNING! {} id has been found more than once! only first one has been loaded!".format(uid))
            continue
        found_uids.add(uid)
        yield uid, po_file


def list_po_dir(root_path, settings):
    """
    Generator. List given directory (expecting one sub-directory per languages)
    and return all files matching languages listed in settings.

    Yield tuples (can_use, uid, num_id, name, isocode, po_path)

    Note that po_path may not actually exists.
    """
    isocodes = ((e, os.path.join(root_path, e, e + ".po")) for e in os.listdir(root_path))
    isocodes = dict(e for e in isocodes if os.path.isfile(e[1]))
    for num_id, name, uid in settings.LANGUAGES[2:]:  # Skip "default" and "en" languages!
        best_po = find_best_isocode_matches(uid, isocodes)
        # print(uid, "->", best_po)
        if best_po:
            isocode = best_po[0]
            yield (True, uid, num_id, name, isocode, isocodes[isocode])
        else:
            yielded = False
            language, _1, _2, language_country, language_variant = locale_explode(uid)
            for isocode in (language, language_variant, language_country, uid):
                p = os.path.join(root_path, isocode, isocode + ".po")
                if not os.path.exists(p):
                    yield (True, uid, num_id, name, isocode, p)
                    yielded = True
                    break
            if not yielded:
                yield (False, uid, num_id, name, None, None)


def enable_addons(addons=None, support=None, disable=False, check_only=False):
    """
    Enable (or disable) addons based either on a set of names, or a set of 'support' types.
    Returns the list of all affected addons (as fake modules)!
    If "check_only" is set, no addon will be enabled nor disabled.
    """
    import addon_utils

    try:
        import bpy
    except ModuleNotFoundError:
        print("Could not import bpy, enable_addons must be run from within Blender.")
        return

    if addons is None:
        addons = {}
    if support is None:
        support = {}

    prefs = bpy.context.preferences
    used_addon_module_names = {addon.module for addon in prefs.addons}
    # In case we need to blacklist some add-ons...
    black_list = {}

    ret = [
        mod for mod in addon_utils.modules()
        if (((addons and mod.__name__ in addons) or
             (not addons and addon_utils.module_bl_info(mod)["support"] in support)) and
            (mod.__name__ not in black_list))
    ]

    if not check_only:
        for mod in ret:
            try:
                addon_module_name = mod.__name__
                if disable:
                    if addon_module_name not in used_addon_module_names:
                        continue
                    print("    Disabling module ", addon_module_name)
                    bpy.ops.preferences.addon_disable(module=addon_module_name)
                else:
                    if addon_module_name in used_addon_module_names:
                        continue
                    print("    Enabling module ", addon_module_name)
                    bpy.ops.preferences.addon_enable(module=addon_module_name)
            except Exception as ex:  # XXX TEMP WORKAROUND
                print(ex)

        # XXX There are currently some problems with bpy/rna...
        #     *Very* tricky to solve!
        #     So this is a hack to make all newly added operator visible by
        #     bpy.types.OperatorProperties.__subclasses__()
        for cat in dir(bpy.ops):
            cat = getattr(bpy.ops, cat)
            for op in dir(cat):
                getattr(cat, op).get_rna_type()

    return ret


##### Main Classes #####

class I18nMessage:
    """
    Internal representation of a message.
    """
    __slots__ = ("msgctxt_lines", "msgid_lines", "msgstr_lines", "comment_lines", "is_fuzzy", "is_commented",
                 "settings")

    def __init__(self, msgctxt_lines=None, msgid_lines=None, msgstr_lines=None, comment_lines=None,
                 is_commented=False, is_fuzzy=False, settings=settings):
        self.settings = settings
        self.msgctxt_lines = msgctxt_lines or []
        self.msgid_lines = msgid_lines or []
        self.msgstr_lines = msgstr_lines or []
        self.comment_lines = comment_lines or []
        self.is_fuzzy = is_fuzzy
        self.is_commented = is_commented

    # ~ def __getstate__(self):
        # ~ return {key: getattr(self, key) for key in self.__slots__}

    # ~ def __getstate__(self):
        # ~ return {key: getattr(self, key) for key in self.__slots__}

    def _get_msgctxt(self):
        return "".join(self.msgctxt_lines)

    def _set_msgctxt(self, ctxt):
        self.msgctxt_lines = [ctxt]
    msgctxt = property(_get_msgctxt, _set_msgctxt)

    def _get_msgid(self):
        return "".join(self.msgid_lines)

    def _set_msgid(self, msgid):
        self.msgid_lines = [msgid]
    msgid = property(_get_msgid, _set_msgid)

    def _get_msgstr(self):
        return "".join(self.msgstr_lines)

    def _set_msgstr(self, msgstr):
        self.msgstr_lines = [msgstr]
    msgstr = property(_get_msgstr, _set_msgstr)

    def _get_sources(self):
        lstrip1 = len(self.settings.PO_COMMENT_PREFIX_SOURCE)
        lstrip2 = len(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM)
        return ([l[lstrip1:] for l in self.comment_lines if l.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE)] +
                [l[lstrip2:] for l in self.comment_lines
                 if l.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM)])

    def _set_sources(self, sources):
        cmmlines = self.comment_lines.copy()
        for l in cmmlines:
            if (
                    l.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE) or
                    l.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM)
            ):
                self.comment_lines.remove(l)
        lines_src = []
        lines_src_custom = []
        for src in sources:
            if is_valid_po_path(src):
                lines_src.append(self.settings.PO_COMMENT_PREFIX_SOURCE + src)
            else:
                lines_src_custom.append(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM + src)
        self.comment_lines += lines_src_custom + lines_src
    sources = property(_get_sources, _set_sources)

    def _get_is_tooltip(self):
        # XXX For now, we assume that all messages > 30 chars are tool-tips!
        return len(self.msgid) > 30
    is_tooltip = property(_get_is_tooltip)

    def copy(self):
        # Deep-copy everything but the settings!
        return self.__class__(msgctxt_lines=self.msgctxt_lines[:], msgid_lines=self.msgid_lines[:],
                              msgstr_lines=self.msgstr_lines[:], comment_lines=self.comment_lines[:],
                              is_commented=self.is_commented, is_fuzzy=self.is_fuzzy, settings=self.settings)

    def normalize(self, max_len=80):
        """
        Normalize this message, call this before exporting it...
        Currently normalize msgctxt, msgid and msgstr lines to given max_len (if below 1, make them single line).
        """
        max_len -= 2  # The two quotes!

        def _splitlines(text):
            lns = text.splitlines()
            return [l + "\n" for l in lns[:-1]] + lns[-1:]

        # We do not need the full power of text-wrap... We just split first at escaped new lines, then into each line
        # if needed... No word splitting, nor fancy spaces handling!
        def _wrap(text, max_len, init_len):
            if len(text) + init_len < max_len:
                return [text]
            lines = _splitlines(text)
            ret = []
            for l in lines:
                tmp = []
                cur_len = 0
                words = l.split(' ')
                for w in words:
                    cur_len += len(w) + 1
                    if cur_len > (max_len - 1) and tmp:
                        ret.append(" ".join(tmp) + " ")
                        del tmp[:]
                        cur_len = len(w) + 1
                    tmp.append(w)
                if tmp:
                    ret.append(" ".join(tmp))
            return ret

        if max_len < 1:
            self.msgctxt_lines = _splitlines(self.msgctxt)
            self.msgid_lines = _splitlines(self.msgid)
            self.msgstr_lines = _splitlines(self.msgstr)
        else:
            init_len = len(self.settings.PO_MSGCTXT) + 1
            if self.is_commented:
                init_len += len(self.settings.PO_COMMENT_PREFIX_MSG)
            self.msgctxt_lines = _wrap(self.msgctxt, max_len, init_len)

            init_len = len(self.settings.PO_MSGID) + 1
            if self.is_commented:
                init_len += len(self.settings.PO_COMMENT_PREFIX_MSG)
            self.msgid_lines = _wrap(self.msgid, max_len, init_len)

            init_len = len(self.settings.PO_MSGSTR) + 1
            if self.is_commented:
                init_len += len(self.settings.PO_COMMENT_PREFIX_MSG)
            self.msgstr_lines = _wrap(self.msgstr, max_len, init_len)

        # Be sure comment lines are not duplicated (can happen with sources...).
        tmp = []
        for l in self.comment_lines:
            if l not in tmp:
                tmp.append(l)
        self.comment_lines = tmp

    _esc_quotes = re.compile(r'(?!<\\)((?:\\\\)*)"')
    _unesc_quotes = re.compile(r'(?!<\\)((?:\\\\)*)\\"')
    _esc_names = ("msgctxt_lines", "msgid_lines", "msgstr_lines")
    _esc_names_all = _esc_names + ("comment_lines",)

    @classmethod
    def do_escape(cls, txt):
        """Replace some chars by their escaped versions!"""
        if "\n" in txt:
            txt = txt.replace("\n", r"\n")
        if "\t" in txt:
            txt.replace("\t", r"\t")
        if '"' in txt:
            txt = cls._esc_quotes.sub(r'\1\"', txt)
        return txt

    @classmethod
    def do_unescape(cls, txt):
        """Replace escaped chars by real ones!"""
        if r"\n" in txt:
            txt = txt.replace(r"\n", "\n")
        if r"\t" in txt:
            txt = txt.replace(r"\t", "\t")
        if r'\"' in txt:
            txt = cls._unesc_quotes.sub(r'\1"', txt)
        return txt

    def escape(self, do_all=False):
        names = self._esc_names_all if do_all else self._esc_names
        for name in names:
            setattr(self, name, [self.do_escape(l) for l in getattr(self, name)])

    def unescape(self, do_all=True):
        names = self._esc_names_all if do_all else self._esc_names
        for name in names:
            setattr(self, name, [self.do_unescape(l) for l in getattr(self, name)])


class I18nMessages:
    """
    Internal representation of messages for one language (iso code), with additional stats info.
    """

    # Avoid parsing again!
    # Keys should be (pseudo) file-names, values are tuples (hash, I18nMessages)
    # Note: only used by po parser currently!
    # _parser_cache = {}

    def __init__(self, uid=None, kind=None, key=None, src=None, settings=settings):
        self.settings = settings
        self.uid = uid if uid is not None else settings.PARSER_TEMPLATE_ID
        self.msgs = self._new_messages()
        self.trans_msgs = set()
        self.fuzzy_msgs = set()
        self.comm_msgs = set()
        self.ttip_msgs = set()
        self.contexts = set()
        self.nbr_msgs = 0
        self.nbr_trans_msgs = 0
        self.nbr_ttips = 0
        self.nbr_trans_ttips = 0
        self.nbr_comm_msgs = 0
        self.nbr_signs = 0
        self.nbr_trans_signs = 0
        self.parsing_errors = []
        if kind and src:
            self.parse(kind, key, src)
        self.update_info()

        self._reverse_cache = None

    def __getstate__(self):
        return (self.settings, self.uid, self.msgs, self.parsing_errors)

    def __setstate__(self, data):
        self.__init__()
        self.settings, self.uid, self.msgs, self.parsing_errors = data
        self.update_info()

    @staticmethod
    def _new_messages():
        return getattr(collections, "OrderedDict", dict)()

    @classmethod
    def gen_empty_messages(cls, uid, blender_ver, blender_hash, bl_time, default_copyright=True, settings=settings):
        """Generate an empty I18nMessages object (only header is present!)."""
        fmt = settings.PO_HEADER_MSGSTR
        msgstr = fmt.format(
            blender_ver=str(blender_ver),
            blender_hash=blender_hash,
            time=time.strftime("%Y-%m-%d %H:%M%z", bl_time),
            uid=str(uid))
        comment = ""
        if default_copyright:
            comment = settings.PO_HEADER_COMMENT_COPYRIGHT.format(year=str(time.gmtime().tm_year))
        comment = comment + settings.PO_HEADER_COMMENT

        msgs = cls(uid=uid, settings=settings)
        key = settings.PO_HEADER_KEY
        msgs.msgs[key] = I18nMessage([key[0]], [key[1]], msgstr.split("\n"), comment.split("\n"),
                                     False, False, settings=settings)
        msgs.update_info()

        return msgs

    def normalize(self, max_len=80):
        for msg in self.msgs.values():
            msg.normalize(max_len)

    def escape(self, do_all=False):
        for msg in self.msgs.values():
            msg.escape(do_all)

    def unescape(self, do_all=True):
        for msg in self.msgs.values():
            msg.unescape(do_all)

    def check(self, fix=False):
        """
        Check consistency between messages and their keys!
        Check messages using format stuff are consistent between msgid and msgstr!
        If fix is True, tries to fix the issues.
        Return a list of found errors (empty if everything went OK!).
        """
        ret = []
        default_context = self.settings.DEFAULT_CONTEXT
        _format = re.compile(self.settings.CHECK_PRINTF_FORMAT).findall
        done_keys = set()
        rem = set()
        tmp = {}
        for key, msg in self.msgs.items():
            msgctxt, msgid, msgstr = msg.msgctxt, msg.msgid, msg.msgstr
            real_key = (msgctxt or default_context, msgid)
            if key != real_key:
                ret.append("Error! msg's context/message do not match its key ({} / {})".format(real_key, key))
                if real_key in self.msgs:
                    ret.append("Error! msg's real_key already used!")
                    if fix:
                        rem.add(real_key)
                elif fix:
                    tmp[real_key] = msg
            done_keys.add(key)
            if '%' in msgid and msgstr and _format(msgid) != _format(msgstr):
                if not msg.is_fuzzy:
                    ret.append("Error! msg's format entities are not matched in msgid and msgstr ({} / \"{}\")"
                               "".format(real_key, msgstr))
                if fix:
                    msg.msgstr = ""
        for k in rem:
            del self.msgs[k]
        self.msgs.update(tmp)
        return ret

    def clean_commented(self):
        self.update_info()
        nbr = len(self.comm_msgs)
        for k in self.comm_msgs:
            del self.msgs[k]
        return nbr

    def rtl_process(self):
        keys = []
        trans = []
        for k, m in self.msgs.items():
            keys.append(k)
            trans.append(m.msgstr)
        trans = utils_rtl.log2vis(trans, self.settings)
        for k, t in zip(keys, trans):
            self.msgs[k].msgstr = t

    def merge(self, msgs, replace=False):
        """
        Merge translations from msgs into self, following those rules:
            * If a msg is in self and not in msgs, keep self untouched.
            * If a msg is in msgs and not in self, skip it.
            * Else (msg both in self and msgs):
                * If self is not translated and msgs is translated or fuzzy, replace by msgs.
                * If self is fuzzy, and msgs is translated, replace by msgs.
                * If self is fuzzy, and msgs is fuzzy, and replace is True, replace by msgs.
                * If self is translated, and msgs is translated, and replace is True, replace by msgs.
                * Else, skip it!
        """
        for k, m in msgs.msgs.items():
            if k not in self.msgs:
                continue
            sm = self.msgs[k]
            if (sm.is_commented or m.is_commented or not m.msgstr):
                continue
            if (not sm.msgstr or replace or (sm.is_fuzzy and (not m.is_fuzzy or replace))):
                sm.msgstr = m.msgstr
                sm.is_fuzzy = m.is_fuzzy
                sm.comment_lines = m.comment_lines

    def update(self, ref, use_similar=None, keep_old_commented=True):
        """
        Update this I18nMessage with the ref one. Translations from ref are never used. Source comments from ref
        completely replace current ones. If use_similar is not 0.0, it will try to match new messages in ref with an
        existing one. Messages no more found in ref will be marked as commented if keep_old_commented is True,
        or removed.
        """
        if use_similar is None:
            use_similar = self.settings.SIMILAR_MSGID_THRESHOLD

        similar_pool = {}
        if use_similar > 0.0:
            for key, msg in self.msgs.items():
                if msg.msgstr:  # No need to waste time with void translations!
                    similar_pool.setdefault(key[1], set()).add(key)

        msgs = self._new_messages().fromkeys(ref.msgs.keys())
        ref_keys = set(ref.msgs.keys())
        org_keys = set(self.msgs.keys())
        new_keys = ref_keys - org_keys
        removed_keys = org_keys - ref_keys

        # First process keys present in both org and ref messages.
        for key in ref_keys - new_keys:
            msg, refmsg = self.msgs[key], ref.msgs[key]
            msg.sources = refmsg.sources
            msg.is_commented = refmsg.is_commented
            msgs[key] = msg

        # Next process new keys.
        if use_similar > 0.0:
            for key, msgid in map(get_best_similar,
                                  tuple((nk, use_similar, tuple(similar_pool.keys())) for nk in new_keys)):
                if msgid:
                    # Try to get the same context, else just get one...
                    skey = (key[0], msgid)
                    if skey not in similar_pool[msgid]:
                        skey = tuple(similar_pool[msgid])[0]
                    # We keep org translation and comments, and mark message as fuzzy.
                    msg, refmsg = self.msgs[skey].copy(), ref.msgs[key].copy()
                    msg.msgctxt = refmsg.msgctxt
                    msg.msgid = refmsg.msgid
                    msg.sources = refmsg.sources
                    msg.is_fuzzy = True
                    msg.is_commented = refmsg.is_commented
                    msgs[key] = msg
                else:
                    msgs[key] = ref.msgs[key].copy()
        else:
            for key in new_keys:
                msgs[key] = ref.msgs[key].copy()

        # Add back all "old" and already commented messages as commented ones, if required
        # (and translation was not void!).
        if keep_old_commented:
            for key in removed_keys:
                msgs[key] = self.msgs[key]
                msgs[key].is_commented = True
                msgs[key].sources = []

        # Special 'meta' message, change project ID version and pot creation date...
        key = self.settings.PO_HEADER_KEY
        rep = []
        markers = ("Project-Id-Version:", "POT-Creation-Date:")
        for mrk in markers:
            for rl in ref.msgs[key].msgstr_lines:
                if rl.startswith(mrk):
                    for idx, ml in enumerate(msgs[key].msgstr_lines):
                        if ml.startswith(mrk):
                            rep.append((idx, rl))
        for idx, txt in rep:
            msgs[key].msgstr_lines[idx] = txt

        # And finalize the update!
        self.msgs = msgs

    def update_info(self):
        self.trans_msgs.clear()
        self.fuzzy_msgs.clear()
        self.comm_msgs.clear()
        self.ttip_msgs.clear()
        self.contexts.clear()
        self.nbr_signs = 0
        self.nbr_trans_signs = 0
        for key, msg in self.msgs.items():
            if key == self.settings.PO_HEADER_KEY:
                continue
            if msg.is_commented:
                self.comm_msgs.add(key)
            else:
                if msg.msgstr:
                    self.trans_msgs.add(key)
                if msg.is_fuzzy:
                    self.fuzzy_msgs.add(key)
                if msg.is_tooltip:
                    self.ttip_msgs.add(key)
                self.contexts.add(key[0])
                self.nbr_signs += len(msg.msgid)
                self.nbr_trans_signs += len(msg.msgstr)
        self.nbr_msgs = len(self.msgs)
        self.nbr_trans_msgs = len(self.trans_msgs - self.fuzzy_msgs)
        self.nbr_ttips = len(self.ttip_msgs)
        self.nbr_trans_ttips = len(self.ttip_msgs & (self.trans_msgs - self.fuzzy_msgs))
        self.nbr_comm_msgs = len(self.comm_msgs)

    def print_info(self, prefix="", output=print, print_stats=True, print_errors=True):
        """
        Print out some info about an I18nMessages object.
        """
        lvl = 0.0
        lvl_ttips = 0.0
        lvl_comm = 0.0
        lvl_trans_ttips = 0.0
        lvl_ttips_in_trans = 0.0
        if self.nbr_msgs > 0:
            lvl = float(self.nbr_trans_msgs) / float(self.nbr_msgs)
            lvl_ttips = float(self.nbr_ttips) / float(self.nbr_msgs)
            lvl_comm = float(self.nbr_comm_msgs) / float(self.nbr_msgs + self.nbr_comm_msgs)
        if self.nbr_ttips > 0:
            lvl_trans_ttips = float(self.nbr_trans_ttips) / float(self.nbr_ttips)
        if self.nbr_trans_msgs > 0:
            lvl_ttips_in_trans = float(self.nbr_trans_ttips) / float(self.nbr_trans_msgs)

        lines = []
        if print_stats:
            lines += [
                "",
                "{:>6.1%} done! ({} translated messages over {}).\n"
                "".format(lvl, self.nbr_trans_msgs, self.nbr_msgs),
                "{:>6.1%} of messages are tooltips ({} over {}).\n"
                "".format(lvl_ttips, self.nbr_ttips, self.nbr_msgs),
                "{:>6.1%} of tooltips are translated ({} over {}).\n"
                "".format(lvl_trans_ttips, self.nbr_trans_ttips, self.nbr_ttips),
                "{:>6.1%} of translated messages are tooltips ({} over {}).\n"
                "".format(lvl_ttips_in_trans, self.nbr_trans_ttips, self.nbr_trans_msgs),
                "{:>6.1%} of messages are commented ({} over {}).\n"
                "".format(lvl_comm, self.nbr_comm_msgs, self.nbr_comm_msgs + self.nbr_msgs),
                "This translation is currently made of {} signs.\n".format(self.nbr_trans_signs)
            ]
        if print_errors and self.parsing_errors:
            lines += ["WARNING! Errors during parsing:\n"]
            lines += ["    Around line {}: {}\n".format(line, error) for line, error in self.parsing_errors]
        output(prefix.join(lines))

    def invalidate_reverse_cache(self, rebuild_now=False):
        """
        Invalidate the reverse cache used by find_best_messages_matches.
        """
        self._reverse_cache = None
        if rebuild_now:
            src_to_msg, ctxt_to_msg, msgid_to_msg, msgstr_to_msg = {}, {}, {}, {}
            ctxt_to_msg.setdefault(self.settings.DEFAULT_CONTEXT, set())
            for key, msg in self.msgs.items():
                if msg.is_commented:
                    continue
                ctxt, msgid = key
                ctxt_to_msg.setdefault(ctxt, set()).add(key)
                msgid_to_msg.setdefault(msgid, set()).add(key)
                msgstr_to_msg.setdefault(msg.msgstr, set()).add(key)
                for src in msg.sources:
                    src_to_msg.setdefault(src, set()).add(key)
            self._reverse_cache = (src_to_msg, ctxt_to_msg, msgid_to_msg, msgstr_to_msg)

    def find_best_messages_matches(self, msgs, msgmap, rna_ctxt, rna_struct_name, rna_prop_name, rna_enum_name):
        """
        Try to find the best I18nMessages (i.e. context/msgid pairs) for the given UI messages:
            msgs: an object containing properties listed in msgmap's values.
            msgmap: a dict of various messages to use for search:
                        {"but_label": subdict, "rna_label": subdict, "enum_label": subdict,
                        "but_tip": subdict, "rna_tip": subdict, "enum_tip": subdict}
                    each subdict being like that:
                        {"msgstr": id, "msgid": id, "msg_flags": id, "key": set()}
                  where msgstr and msgid are identifiers of string props in msgs (resp. translated and org message),
                        msg_flags is not used here, and key is a set of matching (msgctxt, msgid) keys for the item.
            The other parameters are about the RNA element from which the strings come from, if it could be determined:
                rna_ctxt: the labels' i18n context.
                rna_struct_name, rna_prop_name, rna_enum_name: should be self-explanatory!
        """
        try:
            import bpy
        except ModuleNotFoundError:
            print("Could not import bpy, find_best_messages_matches must be run from within Blender.")
            return

        # Build helper mappings.
        # Note it's user responsibility to know when to invalidate (and hence force rebuild) this cache!
        if self._reverse_cache is None:
            self.invalidate_reverse_cache(True)
        src_to_msg, ctxt_to_msg, msgid_to_msg, msgstr_to_msg = self._reverse_cache

    #    print(len(src_to_msg), len(ctxt_to_msg), len(msgid_to_msg), len(msgstr_to_msg))

        # Build RNA key.
        src, src_rna, src_enum = bpy.utils.make_rna_paths(rna_struct_name, rna_prop_name, rna_enum_name)
        print("src: ", src_rna, src_enum)

        # Labels.
        elbl = getattr(msgs, msgmap["enum_label"]["msgstr"])
        if elbl:
            # Enum items' labels have no i18n context...
            k = ctxt_to_msg[self.settings.DEFAULT_CONTEXT].copy()
            if elbl in msgid_to_msg:
                k &= msgid_to_msg[elbl]
            elif elbl in msgstr_to_msg:
                k &= msgstr_to_msg[elbl]
            else:
                k = set()
            # We assume if we already have only one key, it's the good one!
            if len(k) > 1 and src_enum in src_to_msg:
                k &= src_to_msg[src_enum]
            msgmap["enum_label"]["key"] = k
        rlbl = getattr(msgs, msgmap["rna_label"]["msgstr"])
        # print("rna label: " + rlbl, rlbl in msgid_to_msg, rlbl in msgstr_to_msg)
        if rlbl:
            k = ctxt_to_msg.get(rna_ctxt, set()).copy()
            if k and rlbl in msgid_to_msg:
                k &= msgid_to_msg[rlbl]
            elif k and rlbl in msgstr_to_msg:
                k &= msgstr_to_msg[rlbl]
            else:
                k = set()
            # We assume if we already have only one key, it's the good one!
            if len(k) > 1 and src_rna in src_to_msg:
                k &= src_to_msg[src_rna]
            msgmap["rna_label"]["key"] = k
        blbl = getattr(msgs, msgmap["but_label"]["msgstr"])
        blbls = [blbl]
        if blbl.endswith(self.settings.NUM_BUTTON_SUFFIX):
            # Num buttons report their label with a trailing ': '...
            blbls.append(blbl[:-len(self.settings.NUM_BUTTON_SUFFIX)])
        print("button label: " + blbl)
        if blbl and elbl not in blbls and (rlbl not in blbls or rna_ctxt != self.settings.DEFAULT_CONTEXT):
            # Always Default context for button label :/
            k = ctxt_to_msg[self.settings.DEFAULT_CONTEXT].copy()
            found = False
            for bl in blbls:
                if bl in msgid_to_msg:
                    k &= msgid_to_msg[bl]
                    found = True
                    break
                elif bl in msgstr_to_msg:
                    k &= msgstr_to_msg[bl]
                    found = True
                    break
            if not found:
                k = set()
            # XXX No need to check against RNA path here, if blabel is different
            #     from rlabel, should not match anyway!
            msgmap["but_label"]["key"] = k

        # Tips (they never have a specific context).
        etip = getattr(msgs, msgmap["enum_tip"]["msgstr"])
        # print("enum tip: " + etip)
        if etip:
            k = ctxt_to_msg[self.settings.DEFAULT_CONTEXT].copy()
            if etip in msgid_to_msg:
                k &= msgid_to_msg[etip]
            elif etip in msgstr_to_msg:
                k &= msgstr_to_msg[etip]
            else:
                k = set()
            # We assume if we already have only one key, it's the good one!
            if len(k) > 1 and src_enum in src_to_msg:
                k &= src_to_msg[src_enum]
            msgmap["enum_tip"]["key"] = k
        rtip = getattr(msgs, msgmap["rna_tip"]["msgstr"])
        # print("rna tip: " + rtip)
        if rtip:
            k = ctxt_to_msg[self.settings.DEFAULT_CONTEXT].copy()
            if k and rtip in msgid_to_msg:
                k &= msgid_to_msg[rtip]
            elif k and rtip in msgstr_to_msg:
                k &= msgstr_to_msg[rtip]
            else:
                k = set()
            # We assume if we already have only one key, it's the good one!
            if len(k) > 1 and src_rna in src_to_msg:
                k &= src_to_msg[src_rna]
            msgmap["rna_tip"]["key"] = k
            # print(k)
        btip = getattr(msgs, msgmap["but_tip"]["msgstr"])
        # print("button tip: " + btip)
        if btip and btip not in {rtip, etip}:
            k = ctxt_to_msg[self.settings.DEFAULT_CONTEXT].copy()
            if btip in msgid_to_msg:
                k &= msgid_to_msg[btip]
            elif btip in msgstr_to_msg:
                k &= msgstr_to_msg[btip]
            else:
                k = set()
            # XXX No need to check against RNA path here, if btip is different from rtip, should not match anyway!
            msgmap["but_tip"]["key"] = k

    def parse(self, kind, key, src):
        del self.parsing_errors[:]
        self.parsers[kind](self, src, key)
        if self.parsing_errors:
            print("{} ({}):".format(key, src))
            self.print_info(print_stats=False)
            print("The parser solved them as well as it could...")
        self.update_info()

    def parse_messages_from_po(self, src, key=None):
        """
        Parse a po file.
        Note: This function will silently "arrange" mis-formatted entries, thus using afterward write_messages() should
              always produce a po-valid file, though not correct!
        """
        reading_msgid = False
        reading_msgstr = False
        reading_msgctxt = False
        reading_comment = False
        is_commented = False
        is_fuzzy = False
        msgctxt_lines = []
        msgid_lines = []
        msgstr_lines = []
        comment_lines = []

        default_context = self.settings.DEFAULT_CONTEXT

        # Helper function
        def finalize_message(self, line_nr):
            nonlocal reading_msgid, reading_msgstr, reading_msgctxt, reading_comment
            nonlocal is_commented, is_fuzzy, msgid_lines, msgstr_lines, msgctxt_lines, comment_lines

            msgid = I18nMessage.do_unescape("".join(msgid_lines))
            msgctxt = I18nMessage.do_unescape("".join(msgctxt_lines))
            msgkey = (msgctxt or default_context, msgid)

            # Never allow overriding existing msgid/msgctxt pairs!
            if msgkey in self.msgs:
                self.parsing_errors.append((line_nr, "{} context/msgid is already in current messages!".format(msgkey)))
                return

            self.msgs[msgkey] = I18nMessage(msgctxt_lines, msgid_lines, msgstr_lines, comment_lines,
                                            is_commented, is_fuzzy, settings=self.settings)

            # Let's clean up and get ready for next message!
            reading_msgid = reading_msgstr = reading_msgctxt = reading_comment = False
            is_commented = is_fuzzy = False
            msgctxt_lines = []
            msgid_lines = []
            msgstr_lines = []
            comment_lines = []

        # try to use src as file name...
        if os.path.isfile(src):
            if os.stat(src).st_size > self.settings.PARSER_MAX_FILE_SIZE:
                # Security, else we could read arbitrary huge files!
                print("WARNING: skipping file {}, too huge!".format(src))
                return
            if not key:
                key = src
            with open(src, 'r', encoding="utf-8") as f:
                src = f.read()

        _msgctxt = self.settings.PO_MSGCTXT
        _comm_msgctxt = self.settings.PO_COMMENT_PREFIX_MSG + _msgctxt
        _len_msgctxt = len(_msgctxt + '"')
        _len_comm_msgctxt = len(_comm_msgctxt + '"')
        _msgid = self.settings.PO_MSGID
        _comm_msgid = self.settings.PO_COMMENT_PREFIX_MSG + _msgid
        _len_msgid = len(_msgid + '"')
        _len_comm_msgid = len(_comm_msgid + '"')
        _msgstr = self.settings.PO_MSGSTR
        _comm_msgstr = self.settings.PO_COMMENT_PREFIX_MSG + _msgstr
        _len_msgstr = len(_msgstr + '"')
        _len_comm_msgstr = len(_comm_msgstr + '"')
        _comm_str = self.settings.PO_COMMENT_PREFIX_MSG
        _comm_fuzzy = self.settings.PO_COMMENT_FUZZY
        _len_comm_str = len(_comm_str + '"')

        # Main loop over all lines in src...
        for line_nr, line in enumerate(src.splitlines()):
            if line == "":
                if reading_msgstr:
                    finalize_message(self, line_nr)
                continue

            elif line.startswith(_msgctxt) or line.startswith(_comm_msgctxt):
                reading_comment = False
                reading_ctxt = True
                if line.startswith(_comm_str):
                    is_commented = True
                    line = line[_len_comm_msgctxt:-1]
                else:
                    line = line[_len_msgctxt:-1]
                msgctxt_lines.append(line)

            elif line.startswith(_msgid) or line.startswith(_comm_msgid):
                reading_comment = False
                reading_msgid = True
                if line.startswith(_comm_str):
                    if not is_commented and reading_ctxt:
                        self.parsing_errors.append((line_nr, "commented msgid following regular msgctxt"))
                    is_commented = True
                    line = line[_len_comm_msgid:-1]
                else:
                    line = line[_len_msgid:-1]
                reading_ctxt = False
                msgid_lines.append(line)

            elif line.startswith(_msgstr) or line.startswith(_comm_msgstr):
                if not reading_msgid:
                    self.parsing_errors.append((line_nr, "msgstr without a prior msgid"))
                else:
                    reading_msgid = False
                reading_msgstr = True
                if line.startswith(_comm_str):
                    line = line[_len_comm_msgstr:-1]
                    if not is_commented:
                        self.parsing_errors.append((line_nr, "commented msgstr following regular msgid"))
                else:
                    line = line[_len_msgstr:-1]
                    if is_commented:
                        self.parsing_errors.append((line_nr, "regular msgstr following commented msgid"))
                msgstr_lines.append(line)

            elif line.startswith(_comm_str[0]):
                if line.startswith(_comm_str):
                    if reading_msgctxt:
                        if is_commented:
                            msgctxt_lines.append(line[_len_comm_str:-1])
                        else:
                            msgctxt_lines.append(line)
                            self.parsing_errors.append((line_nr, "commented string while reading regular msgctxt"))
                    elif reading_msgid:
                        if is_commented:
                            msgid_lines.append(line[_len_comm_str:-1])
                        else:
                            msgid_lines.append(line)
                            self.parsing_errors.append((line_nr, "commented string while reading regular msgid"))
                    elif reading_msgstr:
                        if is_commented:
                            msgstr_lines.append(line[_len_comm_str:-1])
                        else:
                            msgstr_lines.append(line)
                            self.parsing_errors.append((line_nr, "commented string while reading regular msgstr"))
                else:
                    if reading_msgctxt or reading_msgid or reading_msgstr:
                        self.parsing_errors.append((line_nr,
                                                    "commented string within msgctxt, msgid or msgstr scope, ignored"))
                    elif line.startswith(_comm_fuzzy):
                        is_fuzzy = True
                    else:
                        comment_lines.append(line)
                    reading_comment = True

            else:
                if reading_msgctxt:
                    msgctxt_lines.append(line[1:-1])
                elif reading_msgid:
                    msgid_lines.append(line[1:-1])
                elif reading_msgstr:
                    line = line[1:-1]
                    msgstr_lines.append(line)
                else:
                    self.parsing_errors.append((line_nr, "regular string outside msgctxt, msgid or msgstr scope"))
                    # self.parsing_errors.append((
                    #     line_nr, "regular string outside msgctxt, msgid or msgstr scope:\n\t\t{}"
                    #     "\n\tComments:{}\n\tContext:{}\n\tKey:{}\n\tMessage:{}".format(
                    #         line, comment_lines, msgctxt_lines, msgid_lines, msgstr_lines)))

        # If no final empty line, last message is not finalized!
        if reading_msgstr:
            finalize_message(self, line_nr)
        self.unescape()

    def write(self, kind, dest):
        return self.writers[kind](self, dest)

    def write_messages_to_po(self, fname, compact=False):
        """
        Write messages in fname po file.
        """
        default_context = self.settings.DEFAULT_CONTEXT

        def _write(self, f, compact):
            _msgctxt = self.settings.PO_MSGCTXT
            _msgid = self.settings.PO_MSGID
            _msgstr = self.settings.PO_MSGSTR
            _comm = self.settings.PO_COMMENT_PREFIX_MSG

            self.escape()

            for num, msg in enumerate(self.msgs.values()):
                if compact and (msg.is_commented or msg.is_fuzzy or not msg.msgstr_lines):
                    continue
                if not compact:
                    f.write("\n".join(msg.comment_lines))
                # Only mark as fuzzy if msgstr is not empty!
                if msg.is_fuzzy and msg.msgstr_lines:
                    f.write("\n" + self.settings.PO_COMMENT_FUZZY)
                _p = _comm if msg.is_commented else ""
                chunks = []
                if msg.msgctxt and msg.msgctxt != default_context:
                    if len(msg.msgctxt_lines) > 1:
                        chunks += [
                            "\n" + _p + _msgctxt + "\"\"\n" + _p + "\"",
                            ("\"\n" + _p + "\"").join(msg.msgctxt_lines),
                            "\"",
                        ]
                    else:
                        chunks += ["\n" + _p + _msgctxt + "\"" + msg.msgctxt + "\""]
                if len(msg.msgid_lines) > 1:
                    chunks += [
                        "\n" + _p + _msgid + "\"\"\n" + _p + "\"",
                        ("\"\n" + _p + "\"").join(msg.msgid_lines),
                        "\"",
                    ]
                else:
                    chunks += ["\n" + _p + _msgid + "\"" + msg.msgid + "\""]
                if len(msg.msgstr_lines) > 1:
                    chunks += [
                        "\n" + _p + _msgstr + "\"\"\n" + _p + "\"",
                        ("\"\n" + _p + "\"").join(msg.msgstr_lines),
                        "\"",
                    ]
                else:
                    chunks += ["\n" + _p + _msgstr + "\"" + msg.msgstr + "\""]
                chunks += ["\n\n"]
                f.write("".join(chunks))

            self.unescape()

        self.normalize(max_len=0)  # No wrapping for now...
        if isinstance(fname, str):
            with open(fname, 'w', encoding="utf-8") as f:
                _write(self, f, compact)
        # Else assume fname is already a file(like) object!
        else:
            _write(self, fname, compact)

    def write_messages_to_mo(self, fname):
        """
        Write messages in fname mo file.
        """
        # XXX Temp solution, until I can make own mo generator working...
        import subprocess
        with tempfile.NamedTemporaryFile(mode='w+', encoding="utf-8") as tmp_po_f:
            os.makedirs(os.path.dirname(fname), exist_ok=True)
            self.write_messages_to_po(tmp_po_f)
            cmd = (
                self.settings.GETTEXT_MSGFMT_EXECUTABLE,
                "--statistics",  # show stats
                tmp_po_f.name,
                "-o",
                fname,
            )
            ret = subprocess.run(cmd, capture_output=True)
            return ret
        # XXX Code below is currently broken (generates corrupted mo files it seems :( )!
        # Using http://www.gnu.org/software/gettext/manual/html_node/MO-Files.html notation.
        # Not generating hash table!
        # Only translated, unfuzzy messages are taken into account!
        default_context = self.settings.DEFAULT_CONTEXT
        msgs = tuple(v for v in self.msgs.values() if not (v.is_fuzzy or v.is_commented) and v.msgstr and v.msgid)
        msgs = sorted(msgs[:2],
                      key=lambda e: (e.msgctxt + e.msgid) if (e.msgctxt and e.msgctxt != default_context) else e.msgid)
        magic_nbr = 0x950412de
        format_rev = 0
        N = len(msgs)
        O = 32
        T = O + N * 8
        S = 0
        H = T + N * 8
        # Prepare our data! we need key (optional context and msgid), translation, and offset and length of both.
        # Offset are relative to start of their own list.
        EOT = b"0x04"  # Used to concatenate context and msgid
        _msgid_offset = 0
        _msgstr_offset = 0

        def _gen(v):
            nonlocal _msgid_offset, _msgstr_offset
            msgid = v.msgid.encode("utf-8")
            msgstr = v.msgstr.encode("utf-8")
            if v.msgctxt and v.msgctxt != default_context:
                msgctxt = v.msgctxt.encode("utf-8")
                msgid = msgctxt + EOT + msgid
            # Don't forget the final NULL char!
            _msgid_len = len(msgid) + 1
            _msgstr_len = len(msgstr) + 1
            ret = ((msgid, _msgid_len, _msgid_offset), (msgstr, _msgstr_len, _msgstr_offset))
            _msgid_offset += _msgid_len
            _msgstr_offset += _msgstr_len
            return ret
        msgs = tuple(_gen(v) for v in msgs)
        msgid_start = H
        msgstr_start = msgid_start + _msgid_offset
        print(N, msgstr_start + _msgstr_offset)
        print(msgs)

        with open(fname, 'wb') as f:
            # Header...
            f.write(struct.pack("=8I", magic_nbr, format_rev, N, O, T, S, H, 0))
            # Msgid's length and offset.
            f.write(b"".join(struct.pack("=2I", length, msgid_start + offset) for (_1, length, offset), _2 in msgs))
            # Msgstr's length and offset.
            f.write(b"".join(struct.pack("=2I", length, msgstr_start + offset) for _1, (_2, length, offset) in msgs))
            # No hash table!
            # Msgid's.
            f.write(b"\0".join(msgid for (msgid, _1, _2), _3 in msgs) + b"\0")
            # Msgstr's.
            f.write(b"\0".join(msgstr for _1, (msgstr, _2, _3) in msgs) + b"\0")

    parsers = {
        "PO": parse_messages_from_po,
    }

    writers = {
        "PO": write_messages_to_po,
        "PO_COMPACT": lambda s, fn: s.write_messages_to_po(fn, True),
        "MO": write_messages_to_mo,
    }

    @classmethod
    def update_from_pot_callback(cls, pot, lng, settings):
        """
        Update or create a single PO file (specified by a filepath) from the given POT `I18nMessages` data.

        Callback usable in a context where Blender specific modules (like ``bpy``) are not available.
        """
        import sys
        sys.stdout.reconfigure(encoding="utf-8")
        sys.stderr.reconfigure(encoding="utf-8")
        if not lng['use']:
            return
        if os.path.isfile(lng['po_path']):
            po = cls(uid=lng['uid'], kind='PO', src=lng['po_path'], settings=settings)
            po.update(pot)
        else:
            po = pot
        po.write(kind="PO", dest=lng['po_path'])
        print("{} PO written!".format(lng['uid']))

    @classmethod
    def cleanup_callback(cls, lng, settings):
        """
        Cleanup a single PO file (specified by a filepath).

        Callback usable in a context where Blender specific modules (like ``bpy``) are not available.
        """
        import sys
        sys.stdout.reconfigure(encoding="utf-8")
        sys.stderr.reconfigure(encoding="utf-8")
        if not lng['use']:
            return
        po = cls(uid=lng['uid'], kind='PO', src=lng['po_path'], settings=settings)
        errs = po.check(fix=True)
        cleanedup_commented = po.clean_commented()
        po.write(kind="PO", dest=lng['po_path'])
        print("Processing {} language ({}).\n"
              "Cleaned up {} commented messages.\n".format(lng['name'], lng['uid'], cleanedup_commented) +
              ("Errors in this po, solved as best as possible!\n\t" + "\n\t".join(errs) if errs else "") + "\n")

    @classmethod
    def update_to_blender_repo_callback(cls, lng, settings):
        """
        Cleanup and write a single PO file (specified by a filepath) into the relevant Blender source 'compact' PO file.

        Callback usable in a context where Blender specific modules (like ``bpy``) are not available.
        """
        import sys
        sys.stdout.reconfigure(encoding="utf-8")
        sys.stderr.reconfigure(encoding="utf-8")
        reports = []
        if lng['uid'] in settings.IMPORT_LANGUAGES_SKIP:
            reports.append(
                "Skipping {} language ({}), edit settings if you want to enable it.".format(
                    lng['name'], lng['uid']))
            return lng['uid'], 0.0, reports
        if not lng['use']:
            reports.append("Skipping {} language ({}).".format(lng['name'], lng['uid']))
            return lng['uid'], 0.0, reports
        po = cls(uid=lng['uid'], kind='PO', src=lng['po_path'], settings=settings)
        errs = po.check(fix=True)
        reports.append("Processing {} language ({}).\n"
                       "Cleaned up {} commented messages.\n".format(lng['name'], lng['uid'], po.clean_commented()) +
                       ("Errors in this po, solved as best as possible!\n\t" + "\n\t".join(errs) if errs else ""))
        if lng['uid'] in settings.IMPORT_LANGUAGES_RTL:
            if platform.system() not in {'Linux'}:
                reports.append("Skipping RtL processing of {} language ({}), "
                               "this is only supported on Linux currently.".format(lng['name'], lng['uid']))
            else:
                po.rtl_process()
        po.write(kind="PO_COMPACT", dest=lng['po_path_blender'])
        po.update_info()
        return lng['uid'], po.nbr_trans_msgs / po.nbr_msgs, reports


class I18n:
    """
    Internal representation of a whole translation set.
    """

    @staticmethod
    def _parser_check_file(path, maxsize=settings.PARSER_MAX_FILE_SIZE,
                           _begin_marker=settings.PARSER_PY_MARKER_BEGIN,
                           _end_marker=settings.PARSER_PY_MARKER_END):
        if os.stat(path).st_size > maxsize:
            # Security, else we could read arbitrary huge files!
            print("WARNING: skipping file {}, too huge!".format(path))
            return None, None, None, False
        txt = ""
        with open(path, encoding="utf8") as f:
            txt = f.read()
        _in = 0
        _out = len(txt)
        if _begin_marker:
            _in = None
            if _begin_marker in txt:
                _in = txt.index(_begin_marker) + len(_begin_marker)
        if _end_marker:
            _out = None
            if _end_marker in txt:
                _out = txt.index(_end_marker)
        if _in is not None and _out is not None:
            in_txt, txt, out_txt = txt[:_in], txt[_in:_out], txt[_out:]
        elif _in is not None:
            in_txt, txt, out_txt = txt[:_in], txt[_in:], None
        elif _out is not None:
            in_txt, txt, out_txt = None, txt[:_out], txt[_out:]
        else:
            in_txt, txt, out_txt = None, txt, None
        return in_txt, txt, out_txt, (True if "translations_tuple" in txt else False)

    @staticmethod
    def _dst(self, path, uid, kind):
        if isinstance(path, str):
            if kind == 'PO':
                if uid == self.settings.PARSER_TEMPLATE_ID:
                    if not path.endswith(".pot"):
                        return os.path.join(os.path.dirname(path), "blender.pot")
                if not path.endswith(".po"):
                    return os.path.join(os.path.dirname(path), uid + ".po")
            elif kind == 'PY':
                if not path.endswith(".py"):
                    if os.path.isdir(path):
                        return os.path.join(path, "translations.py")
                    return os.path.join(os.path.dirname(path), "translations.py")
        return path

    def __init__(self, kind=None, src=None, langs=set(), settings=settings):
        self.settings = settings
        self.trans: dict[str, I18nMessages] = {}
        self.src = {}  # Should have the same keys as self.trans (plus PARSER_PY_ID for py file)!
        self.dst = self._dst  # A callable that transforms src_path into dst_path!
        if kind and src:
            self.parse(kind, src, langs)
        self.update_info()

    def _py_file_get(self):
        return self.src.get(self.settings.PARSER_PY_ID)

    def _py_file_set(self, value):
        self.src[self.settings.PARSER_PY_ID] = value
    py_file = property(_py_file_get, _py_file_set)

    def escape(self, do_all=False):
        for trans in self.trans.values():
            trans.escape(do_all)

    def unescape(self, do_all=True):
        for trans in self.trans.values():
            trans.unescape(do_all)

    def update_info(self):
        self.nbr_trans = 0
        self.lvl = 0.0
        self.lvl_ttips = 0.0
        self.lvl_trans_ttips = 0.0
        self.lvl_ttips_in_trans = 0.0
        self.lvl_comm = 0.0
        self.nbr_signs = 0
        self.nbr_trans_signs = 0
        self.contexts = set()

        if self.settings.PARSER_TEMPLATE_ID in self.trans:
            self.nbr_trans = len(self.trans) - 1
            self.nbr_signs = self.trans[self.settings.PARSER_TEMPLATE_ID].nbr_signs
        else:
            self.nbr_trans = len(self.trans)
        for msgs in self.trans.values():
            msgs.update_info()
            if msgs.nbr_msgs > 0:
                self.lvl += float(msgs.nbr_trans_msgs) / float(msgs.nbr_msgs)
                self.lvl_ttips += float(msgs.nbr_ttips) / float(msgs.nbr_msgs)
                self.lvl_comm += float(msgs.nbr_comm_msgs) / float(msgs.nbr_msgs + msgs.nbr_comm_msgs)
            if msgs.nbr_ttips > 0:
                self.lvl_trans_ttips = float(msgs.nbr_trans_ttips) / float(msgs.nbr_ttips)
            if msgs.nbr_trans_msgs > 0:
                self.lvl_ttips_in_trans = float(msgs.nbr_trans_ttips) / float(msgs.nbr_trans_msgs)
            if self.nbr_signs == 0:
                self.nbr_signs = msgs.nbr_signs
            self.nbr_trans_signs += msgs.nbr_trans_signs
            self.contexts |= msgs.contexts

    def print_stats(self, prefix="", print_msgs=True):
        """
        Print out some stats about an I18n object.
        If print_msgs is True, it will also print all its translations' stats.
        """
        if print_msgs:
            msgs_prefix = prefix + "    "
            for key, msgs in self.trans.items():
                if key == self.settings.PARSER_TEMPLATE_ID:
                    continue
                print(prefix + key + ":")
                msgs.print_stats(prefix=msgs_prefix)
                print(prefix)

        nbr_contexts = len(self.contexts - {self.settings.DEFAULT_CONTEXT})
        if nbr_contexts != 1:
            if nbr_contexts == 0:
                nbr_contexts = "No"
            _ctx_txt = "s are"
        else:
            _ctx_txt = " is"
        lines = ((
            "",
            "Average stats for all {} translations:\n".format(self.nbr_trans),
            "    {:>6.1%} done!\n".format(self.lvl / self.nbr_trans),
            "    {:>6.1%} of messages are tooltips.\n".format(self.lvl_ttips / self.nbr_trans),
            "    {:>6.1%} of tooltips are translated.\n".format(self.lvl_trans_ttips / self.nbr_trans),
            "    {:>6.1%} of translated messages are tooltips.\n".format(self.lvl_ttips_in_trans / self.nbr_trans),
            "    {:>6.1%} of messages are commented.\n".format(self.lvl_comm / self.nbr_trans),
            "    The org msgids are currently made of {} signs.\n".format(self.nbr_signs),
            "    All processed translations are currently made of {} signs.\n".format(self.nbr_trans_signs),
            "    {} specific context{} present:\n".format(self.nbr_contexts, _ctx_txt)) +
            tuple("            " + c + "\n" for c in self.contexts - {self.settings.DEFAULT_CONTEXT}) +
            ("\n",)
        )
        print(prefix.join(lines))

    @classmethod
    def check_py_module_has_translations(cls, src, settings=settings):
        """
        Check whether a given src (a py module, either a directory or a py file) has some i18n translation data,
        and returns a tuple (src_file, translations_tuple) if yes, else (None, None).
        """
        txts = []
        if os.path.isdir(src):
            for root, _dnames, fnames in os.walk(src):
                for fname in fnames:
                    if not fname.endswith(".py"):
                        continue
                    path = os.path.join(root, fname)
                    _1, txt, _2, has_trans = cls._parser_check_file(path)
                    if has_trans:
                        txts.append((path, txt))
        elif src.endswith(".py") and os.path.isfile(src):
            _1, txt, _2, has_trans = cls._parser_check_file(src)
            if has_trans:
                txts.append((src, txt))
        for path, txt in txts:
            tuple_id = "translations_tuple"
            env = globals().copy()
            exec(txt, env)
            if tuple_id in env:
                return path, env[tuple_id]
        return None, None  # No data...

    def parse(self, kind, src, langs=set()):
        self.parsers[kind](self, src, langs)

    def parse_from_po(self, src, langs=set()):
        """
        src must be a tuple (dir_of_pos, pot_file), where:
            * dir_of_pos may either contains iso_CODE.po files, and/or iso_CODE/iso_CODE.po files.
            * pot_file may be None (in which case there will be no ref messages).
        if langs set is void, all languages found are loaded.
        """
        root_dir, pot_file = src
        if pot_file and os.path.isfile(pot_file):
            self.trans[self.settings.PARSER_TEMPLATE_ID] = I18nMessages(self.settings.PARSER_TEMPLATE_ID, 'PO',
                                                                        pot_file, pot_file, settings=self.settings)
            self.src_po[self.settings.PARSER_TEMPLATE_ID] = pot_file

        for uid, po_file in get_po_files_from_dir(root_dir, langs):
            self.trans[uid] = I18nMessages(uid, 'PO', po_file, po_file, settings=self.settings)
            self.src_po[uid] = po_file

    def parse_from_py(self, src, langs=set()):
        """
        src must be a valid path, either a py file or a module directory (in which case all py files inside it
        will be checked, first file matching will win!).
        if langs set is void, all languages found are loaded.
        """
        default_context = self.settings.DEFAULT_CONTEXT
        self.py_file, msgs = self.check_py_module_has_translations(src, self.settings)
        if msgs is None:
            self.py_file = src
            msgs = ()
        for key, (sources, gen_comments), *translations in msgs:
            if self.settings.PARSER_TEMPLATE_ID not in self.trans:
                self.trans[self.settings.PARSER_TEMPLATE_ID] = I18nMessages(self.settings.PARSER_TEMPLATE_ID,
                                                                            settings=self.settings)
                self.src[self.settings.PARSER_TEMPLATE_ID] = self.py_file
            if key in self.trans[self.settings.PARSER_TEMPLATE_ID].msgs:
                print("ERROR! key {} is defined more than once! Skipping re-definitions!")
                continue
            custom_src = [c for c in sources if c.startswith("bpy.")]
            src = [c for c in sources if not c.startswith("bpy.")]
            common_comment_lines = [self.settings.PO_COMMENT_PREFIX_GENERATED + c for c in gen_comments] + \
                                   [self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM + c for c in custom_src] + \
                                   [self.settings.PO_COMMENT_PREFIX_SOURCE + c for c in src]
            ctxt = [key[0]] if key[0] else [default_context]
            self.trans[self.settings.PARSER_TEMPLATE_ID].msgs[key] = I18nMessage(ctxt, [key[1]], [""],
                                                                                 common_comment_lines, False, False,
                                                                                 settings=self.settings)
            for uid, msgstr, (is_fuzzy, user_comments) in translations:
                if uid not in self.trans:
                    self.trans[uid] = I18nMessages(uid, settings=self.settings)
                    self.src[uid] = self.py_file
                comment_lines = [self.settings.PO_COMMENT_PREFIX + c for c in user_comments] + common_comment_lines
                self.trans[uid].msgs[key] = I18nMessage(ctxt, [key[1]], [msgstr], comment_lines, False, is_fuzzy,
                                                        settings=self.settings)
        # key = self.settings.PO_HEADER_KEY
        # for uid, trans in self.trans.items():
        #     if key not in trans.msgs:
        #         trans.msgs[key]
        self.unescape()

    def write(self, kind, langs=set()):
        return self.writers[kind](self, langs)

    def write_to_po(self, langs=set()):
        """
        Write all translations into po files. By default, write in the same files (or dir) as the source, specify
        a custom self.dst function to write somewhere else!
        Note: If langs is set and you want to export the pot template as well, langs must contain PARSER_TEMPLATE_ID
              ({} currently).
        """.format(self.settings.PARSER_TEMPLATE_ID)
        keys = self.trans.keys()
        if langs:
            keys &= langs
        for uid in keys:
            dst = self.dst(self, self.src.get(uid, ""), uid, 'PO')
            self.trans[uid].write('PO', dst)

    def write_to_py(self, langs=set()):
        """
        Write all translations as python code, either in a "translations.py" file under same dir as source(s), or in
        specified file if self.py_file is set (default, as usual can be customized with self.dst callable!).
        Note: If langs is set and you want to export the pot template as well, langs must contain PARSER_TEMPLATE_ID
              ({} currently).
        """.format(self.settings.PARSER_TEMPLATE_ID)
        default_context = self.settings.DEFAULT_CONTEXT

        def _gen_py(self, langs, tab="    "):
            _lencomm = len(self.settings.PO_COMMENT_PREFIX)
            _lengen = len(self.settings.PO_COMMENT_PREFIX_GENERATED)
            _lensrc = len(self.settings.PO_COMMENT_PREFIX_SOURCE)
            _lencsrc = len(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM)
            ret = [
                "# NOTE: You can safely move around this auto-generated block (with the begin/end markers!),",
                "#       and edit the translations by hand.",
                "#       Just carefully respect the format of the tuple!",
                "",
                "# Tuple of tuples:",
                "# ((msgctxt, msgid), (sources, gen_comments), (lang, translation, (is_fuzzy, comments)), ...)",
                "translations_tuple = (",
            ]
            # First gather all keys (msgctxt, msgid) - theoretically, all translations should share the same, but...
            # Note: using an ordered dict if possible (stupid sets cannot be ordered :/ ).
            keys = I18nMessages._new_messages()
            for trans in self.trans.values():
                keys.update(trans.msgs)
            # Get the ref translation (ideally, PARSER_TEMPLATE_ID one, else the first one that pops up!
            # Ref translation will be used to generate sources "comments"
            ref = self.trans.get(self.settings.PARSER_TEMPLATE_ID) or self.trans[list(self.trans.keys())[0]]
            # Get all languages (uids) and sort them (PARSER_TEMPLATE_ID and PARSER_PY_ID excluded!)
            translations = self.trans.keys() - {self.settings.PARSER_TEMPLATE_ID, self.settings.PARSER_PY_ID}
            if langs:
                translations &= langs
            translations = [('"' + lng + '"', self.trans[lng]) for lng in sorted(translations)]
            print(f"Translated {len(keys)} keys saved to .py file:")
            print(*(k for k in keys.keys()), sep="\n")
            for key in keys.keys():
                if ref.msgs[key].is_commented:
                    continue
                # Key (context + msgid).
                msgctxt, msgid = ref.msgs[key].msgctxt, ref.msgs[key].msgid
                if not msgctxt:
                    msgctxt = default_context
                ret.append(tab + "(({}, \"{}\"),".format('"' + msgctxt + '"' if msgctxt else "None", msgid))
                # Common comments (mostly sources!).
                sources = []
                gen_comments = []
                for comment in ref.msgs[key].comment_lines:
                    if comment.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE_CUSTOM):
                        sources.append(comment[_lencsrc:])
                    elif comment.startswith(self.settings.PO_COMMENT_PREFIX_SOURCE):
                        sources.append(comment[_lensrc:])
                    elif comment.startswith(self.settings.PO_COMMENT_PREFIX_GENERATED):
                        gen_comments.append(comment[_lengen:])
                if not (sources or gen_comments):
                    ret.append(tab + " ((), ()),")
                else:
                    if len(sources) > 1:
                        ret.append(tab + ' (("' + sources[0] + '",')
                        ret += [tab + '   "' + s + '",' for s in sources[1:-1]]
                        ret.append(tab + '   "' + sources[-1] + '"),')
                    else:
                        ret.append(tab + " ((" + ('"' + sources[0] + '",' if sources else "") + "),")
                    if len(gen_comments) > 1:
                        ret.append(tab + '  ("' + gen_comments[0] + '",')
                        ret += [tab + '   "' + s + '",' for s in gen_comments[1:-1]]
                        ret.append(tab + '   "' + gen_comments[-1] + '")),')
                    else:
                        ret.append(tab + "  (" + ('"' + gen_comments[0] + '",' if gen_comments else "") + ")),")
                # All languages
                for lngstr, trans in translations:
                    if key not in trans.msgs:
                        continue  # Happens if no language is enabled in translation UI, and new messages are generated.
                    if trans.msgs[key].is_commented:
                        continue
                    # Language code and translation.
                    ret.append(tab + " (" + lngstr + ', "' + trans.msgs[key].msgstr + '",')
                    # User comments and fuzzy.
                    comments = []
                    for comment in trans.msgs[key].comment_lines:
                        if comment.startswith(self.settings.PO_COMMENT_PREFIX):
                            comments.append(comment[_lencomm:])
                    ret.append(tab + "  (" + ("True" if trans.msgs[key].is_fuzzy else "False") + ",")
                    if len(comments) > 1:
                        ret.append(tab + '   ("' + comments[0] + '",')
                        ret += [tab * 2 + '"' + s + '",' for s in comments[1:-1]]
                        ret.append(tab * 2 + '"' + comments[-1] + '"))),')
                    else:
                        ret[-1] = ret[-1] + " (" + (('"' + comments[0] + '",') if comments else "") + "))),"

                ret.append("     " + "),")
            ret += [
                ")",
                "",
                "translations_dict = {}",
                "for msg in translations_tuple:",
                tab + "key = msg[0]",
                tab + "for lang, trans, (is_fuzzy, comments) in msg[2:]:",
                tab * 2 + "if trans and not is_fuzzy:",
                tab * 3 + "translations_dict.setdefault(lang, {})[key] = trans",
                "",
            ]
            return ret

        self.escape(True)
        dst = self.dst(self, self.src.get(self.settings.PARSER_PY_ID, ""), self.settings.PARSER_PY_ID, 'PY')
        print(dst)
        prev = txt = nxt = ""
        if os.path.exists(dst):
            if not os.path.isfile(dst):
                print("WARNING: trying to write as python code into {}, which is not a file! Aborting.".format(dst))
                return
            prev, txt, nxt, _has_trans = self._parser_check_file(dst)
            if prev is None and nxt is None:
                print("WARNING: Looks like given python file {} has no auto-generated translations yet, will be added "
                      "at the end of the file, you can move that section later if needed...".format(dst))
                txt = ([txt, "", self.settings.PARSER_PY_MARKER_BEGIN] +
                       _gen_py(self, langs) +
                       ["", self.settings.PARSER_PY_MARKER_END])
            else:
                # We completely replace the text found between start and end markers...
                txt = _gen_py(self, langs)
        else:
            print("Creating python file {} containing translations.".format(dst))
            txt = [
                "# SPDX-License-Identifier: GPL-2.0-or-later",
                "",
                self.settings.PARSER_PY_MARKER_BEGIN,
                "",
            ]
            txt += _gen_py(self, langs)
            txt += [
                "",
                self.settings.PARSER_PY_MARKER_END,
            ]
        with open(dst, 'w', encoding="utf8") as f:
            f.write((prev or "") + "\n".join(txt) + (nxt or ""))
        self.unescape()

    parsers = {
        "PO": parse_from_po,
        "PY": parse_from_py,
    }

    writers = {
        "PO": write_to_po,
        "PY": write_to_py,
    }



import bpy


class MyAssetShelf(bpy.types.AssetShelf):
    bl_space_type = 'VIEW_3D'
    bl_idname = "VIEW3D_AST_my_asset_shelf"

    @classmethod
    def poll(cls, context):
        return context.mode == 'OBJECT'

    @classmethod
    def asset_poll(cls, asset):
        return asset.id_type in {'MATERIAL', 'OBJECT'}


def register():
    bpy.utils.register_class(MyAssetShelf)


def unregister():
    bpy.utils.unregister_class(MyAssetShelf)


if __name__ == "__main__":
    register()
